@@ -19,27 +19,40 @@
 
 package org.apache.curator.framework;
 
-import com.google.common.collect.ImmutableList;
+import java.net.InetAddress;
+import java.net.UnknownHostException;
+import java.util.Arrays;
+import java.util.List;
+import java.util.Objects;
+import java.util.concurrent.Executor;
+import java.util.concurrent.ThreadFactory;
+import java.util.concurrent.TimeUnit;
+
+import org.apache.curator.CuratorZookeeperClient;
 import org.apache.curator.RetryPolicy;
 import org.apache.curator.ensemble.EnsembleProvider;
 import org.apache.curator.ensemble.fixed.FixedEnsembleProvider;
 import org.apache.curator.framework.api.ACLProvider;
 import org.apache.curator.framework.api.CompressionProvider;
+import org.apache.curator.framework.api.CreateBuilder;
 import org.apache.curator.framework.api.PathAndBytesable;
 import org.apache.curator.framework.imps.CuratorFrameworkImpl;
 import org.apache.curator.framework.imps.CuratorTempFrameworkImpl;
 import org.apache.curator.framework.imps.DefaultACLProvider;
 import org.apache.curator.framework.imps.GzipCompressionProvider;
+import org.apache.curator.framework.schema.SchemaSet;
+import org.apache.curator.framework.state.ConnectionStateErrorPolicy;
+import org.apache.curator.framework.state.ConnectionStateListenerManagerFactory;
+import org.apache.curator.framework.state.StandardConnectionStateErrorPolicy;
 import org.apache.curator.utils.DefaultZookeeperFactory;
 import org.apache.curator.utils.ZookeeperFactory;
+import org.apache.zookeeper.CreateMode;
 import org.apache.zookeeper.Watcher;
 import org.apache.zookeeper.ZooKeeper;
-import java.net.InetAddress;
-import java.net.UnknownHostException;
-import java.util.Arrays;
-import java.util.List;
-import java.util.concurrent.ThreadFactory;
-import java.util.concurrent.TimeUnit;
+import org.apache.zookeeper.client.ZKClientConfig;
+
+import com.google.common.base.Preconditions;
+import com.google.common.collect.ImmutableList;
 
 /**
  * Factory methods for creating framework-style clients
@@ -56,6 +69,7 @@ public class CuratorFrameworkFactory
     private static final DefaultACLProvider DEFAULT_ACL_PROVIDER = new DefaultACLProvider();
     private static final long DEFAULT_INACTIVE_THRESHOLD_MS = (int)TimeUnit.MINUTES.toMillis(3);
     private static final int DEFAULT_CLOSE_WAIT_MS = (int)TimeUnit.SECONDS.toMillis(1);
+    private static final boolean DEFAULT_WITH_ENSEMBLE_TRACKER = true;
 
     /**
      * Return a new builder that builds a CuratorFramework
@@ -97,10 +111,52 @@ public class CuratorFrameworkFactory
             retryPolicy(retryPolicy).
             build();
     }
+    
+    /**
+     * Create a new client
+     *
+     * @param connectString       list of servers to connect to
+     * @param sessionTimeoutMs    session timeout
+     * @param connectionTimeoutMs connection timeout
+     * @param retryPolicy         retry policy to use
+     * @param zkClientConfig      ZKClientConfig
+     * @return client
+     * 
+     * @since 5.1.1, supported from ZooKeeper 3.6.1 and above.
+     */
+    public static CuratorFramework newClient(String connectString, int sessionTimeoutMs, int connectionTimeoutMs, RetryPolicy retryPolicy, ZKClientConfig zkClientConfig)
+    {
+        return builder().
+            connectString(connectString).
+            sessionTimeoutMs(sessionTimeoutMs).
+            connectionTimeoutMs(connectionTimeoutMs).
+            retryPolicy(retryPolicy).
+            zkClientConfig(zkClientConfig).
+            build();
+    }
+
+    /**
+     * Return the local address as bytes that can be used as a node payload
+     *
+     * @return local address bytes
+     */
+    public static byte[] getLocalAddress()
+    {
+        try
+        {
+            return InetAddress.getLocalHost().getHostAddress().getBytes();
+        }
+        catch ( UnknownHostException ignore )
+        {
+            // ignore
+        }
+        return new byte[0];
+    }
 
     public static class Builder
     {
         private EnsembleProvider ensembleProvider;
+        private boolean withEnsembleTracker = DEFAULT_WITH_ENSEMBLE_TRACKER;
         private int sessionTimeoutMs = DEFAULT_SESSION_TIMEOUT_MS;
         private int connectionTimeoutMs = DEFAULT_CONNECTION_TIMEOUT_MS;
         private int maxCloseWaitMs = DEFAULT_CLOSE_WAIT_MS;
@@ -113,6 +169,14 @@ public class CuratorFrameworkFactory
         private ZookeeperFactory zookeeperFactory = DEFAULT_ZOOKEEPER_FACTORY;
         private ACLProvider aclProvider = DEFAULT_ACL_PROVIDER;
         private boolean canBeReadOnly = false;
+        private boolean useContainerParentsIfAvailable = true;
+        private ConnectionStateErrorPolicy connectionStateErrorPolicy = new StandardConnectionStateErrorPolicy();
+        private SchemaSet schemaSet = SchemaSet.getDefaultSchemaSet();
+        private int waitForShutdownTimeoutMs = 0;
+        private Executor runSafeService = null;
+        private ConnectionStateListenerManagerFactory connectionStateListenerManagerFactory = ConnectionStateListenerManagerFactory.standard;
+        private int simulatedSessionExpirationPercent = 100;
+        private ZKClientConfig zkClientConfig;
 
         /**
          * Apply the current values and build a new CuratorFramework
@@ -154,7 +218,7 @@ public class CuratorFrameworkFactory
 
         /**
          * Add connection authorization
-         * 
+         *
          * Subsequent calls to this method overwrite the prior calls.
          *
          * @param scheme the scheme
@@ -207,6 +271,29 @@ public class CuratorFrameworkFactory
             return this;
         }
 
+        /**
+         * Allows to configure if the ensemble configuration changes will be watched.
+         * The default value is {@code true}.<br>
+         * 
+         * IMPORTANT: Use this method in combination with {@link #ensembleProvider(EnsembleProvider)} to provide
+         * an instance that returns {@code false} on {@link EnsembleProvider#updateServerListEnabled()} in order
+         * to fully achieve that ensemble server list changes are ignored<br>
+         * 
+         * @param withEnsembleTracker use {@code false} if you want to avoid following ensemble configuration changes
+         * @return this
+         */
+        public Builder ensembleTracker(boolean withEnsembleTracker) {
+            this.withEnsembleTracker = withEnsembleTracker;
+            return this;
+        }
+
+        /**
+         * @return {@code true} if ensemble configuration changes MUST be watched
+         */
+        public boolean withEnsembleTracker() {
+            return withEnsembleTracker;
+        }
+
         /**
          * Sets the data to use when {@link PathAndBytesable#forPath(String)} is used.
          * This is useful for debugging purposes. For example, you could set this to be the IP of the
@@ -328,6 +415,139 @@ public class CuratorFrameworkFactory
             return this;
         }
 
+        /**
+         * By default, Curator uses {@link CreateBuilder#creatingParentContainersIfNeeded()}
+         * if the ZK JAR supports {@link CreateMode#CONTAINER}. Call this method to turn off this behavior.
+         *
+         * @return this
+         */
+        public Builder dontUseContainerParents()
+        {
+            this.useContainerParentsIfAvailable = false;
+            return this;
+        }
+
+        /**
+         * Set the error policy to use. The default is {@link StandardConnectionStateErrorPolicy}
+         *
+         * @since 3.0.0
+         * @param connectionStateErrorPolicy new error policy
+         * @return this
+         */
+        public Builder connectionStateErrorPolicy(ConnectionStateErrorPolicy connectionStateErrorPolicy)
+        {
+            this.connectionStateErrorPolicy = connectionStateErrorPolicy;
+            return this;
+        }
+
+        /**
+         * Set a timeout for {@link CuratorZookeeperClient#close(int)}  }.
+         * The default is 0, which means that this feature is disabled.
+         *
+         * @since 4.0.2
+         * @param waitForShutdownTimeoutMs default timeout
+         * @return this
+         */
+        public Builder waitForShutdownTimeoutMs(int waitForShutdownTimeoutMs)
+        {
+            this.waitForShutdownTimeoutMs = waitForShutdownTimeoutMs;
+            return this;
+        }
+
+        /**
+         * <p>
+         *     Prior to 3.0.0, Curator did not try to manage session expiration
+         *     other than the functionality provided by ZooKeeper itself. Starting with
+         *     3.0.0, Curator has the option of attempting to monitor session expiration
+         *     above what is provided by ZooKeeper. The percentage set by this method
+         *     determines how and if Curator will check for session expiration.
+         * </p>
+         *
+         * <p>
+         *     The default percentage is 100.
+         * </p>
+         *
+         * <p>
+         *     If it is set to <tt>0</tt>, Curator does not do any additional checking
+         *     for session expiration.
+         * </p>
+         *
+         * <p>
+         *     If a positive number is set, Curator will check for session expiration
+         *     as follows: when ZooKeeper sends a Disconnect event, Curator will start a timer.
+         *     If re-connection is not achieved before the elapsed time exceeds the negotiated
+         *     session time multiplied by the session expiration percent, Curator will simulate
+         *     a session expiration. Due to timing/network issues, it is <b>not possible</b> for
+         *     a client to match the server's session timeout with complete accuracy. Thus, the need
+         *     for a session expiration percentage.
+         * </p>
+         *
+         * @param simulatedSessionExpirationPercent new simulated session expiration percentage
+         * @return this
+         * @since 5.0
+         */
+        public Builder simulatedSessionExpirationPercent(int simulatedSessionExpirationPercent) {
+            Preconditions.checkArgument(
+                    (simulatedSessionExpirationPercent > 0) && (simulatedSessionExpirationPercent <= 100),
+                    "simulatedSessionExpirationPercent must be > 0 and <= 100");
+            this.simulatedSessionExpirationPercent = simulatedSessionExpirationPercent;
+            return this;
+        }
+        
+        public Builder zkClientConfig(ZKClientConfig zkClientConfig) {
+        	this.zkClientConfig = zkClientConfig;
+        	return this;
+        }
+
+        /**
+         * Add an enforced schema set
+         *
+         * @param schemaSet the schema set
+         * @return this
+         * @since 3.2.0
+         */
+        public Builder schemaSet(SchemaSet schemaSet)
+        {
+            this.schemaSet = schemaSet;
+            return this;
+        }
+
+        /**
+         * Curator (and user) recipes will use this executor to call notifyAll
+         * and other blocking calls that might normally block ZooKeeper's event thread.
+         * By default, an executor is allocated internally using the provided (or default)
+         * {@link #threadFactory(java.util.concurrent.ThreadFactory)}. Use this method
+         * to set a custom executor.
+         *
+         * @param runSafeService executor to use for calls to notifyAll from Watcher callbacks etc
+         * @return this
+         * @since 4.1.0
+         */
+        public Builder runSafeService(Executor runSafeService)
+        {
+            this.runSafeService = runSafeService;
+            return this;
+        }
+
+        /**
+         * Sets the connection state listener manager factory. For example,
+         * you can set {@link org.apache.curator.framework.state.ConnectionStateListenerManagerFactory#circuitBreaking(org.apache.curator.RetryPolicy)}
+         *
+         * @param connectionStateListenerManagerFactory manager factory to use
+         * @return this
+         * @since 4.2.0
+         */
+        public Builder connectionStateListenerManagerFactory(ConnectionStateListenerManagerFactory connectionStateListenerManagerFactory)
+        {
+            this.connectionStateListenerManagerFactory = Objects.requireNonNull(connectionStateListenerManagerFactory, "connectionStateListenerManagerFactory cannot be null");
+            return this;
+        }
+
+        public Executor getRunSafeService()
+        {
+            return runSafeService;
+        }
+
         public ACLProvider getAclProvider()
         {
             return aclProvider;
@@ -363,6 +583,11 @@ public class CuratorFrameworkFactory
             return connectionTimeoutMs;
         }
 
+        public int getWaitForShutdownTimeoutMs()
+        {
+            return waitForShutdownTimeoutMs;
+        }
+
         public int getMaxCloseWaitMs()
         {
             return maxCloseWaitMs;
@@ -378,6 +603,29 @@ public class CuratorFrameworkFactory
             return namespace;
         }
 
+        public boolean useContainerParentsIfAvailable()
+        {
+            return useContainerParentsIfAvailable;
+        }
+
+        public ConnectionStateErrorPolicy getConnectionStateErrorPolicy()
+        {
+            return connectionStateErrorPolicy;
+        }
+
+        public int getSimulatedSessionExpirationPercent() {
+            return simulatedSessionExpirationPercent;
+        }
+        
+        public ZKClientConfig getZkClientConfig() {
+            return zkClientConfig;
+        }
+
+        public SchemaSet getSchemaSet()
+        {
+            return schemaSet;
+        }
+
         @Deprecated
         public String getAuthScheme()
         {
@@ -440,24 +688,16 @@ public class CuratorFrameworkFactory
             return canBeReadOnly;
         }
 
+        public ConnectionStateListenerManagerFactory getConnectionStateListenerManagerFactory()
+        {
+            return connectionStateListenerManagerFactory;
+        }
+
         private Builder()
         {
         }
     }
 
-    private static byte[] getLocalAddress()
-    {
-        try
-        {
-            return InetAddress.getLocalHost().getHostAddress().getBytes();
-        }
-        catch ( UnknownHostException ignore )
-        {
-            // ignore
-        }
-        return new byte[0];
-    }
-
     private CuratorFrameworkFactory()
     {
     }
