@@ -23,6 +23,7 @@ import org.apache.zookeeper.data.Stat;
 public interface SetDataBuilder extends
     BackgroundPathAndBytesable<Stat>,
     Versionable<BackgroundPathAndBytesable<Stat>>,
-    Compressible<SetDataBackgroundVersionable>
+    Compressible<SetDataBackgroundVersionable>,
+    Idempotentable<SetDataBuilder>
 {
 }
