@@ -20,19 +20,23 @@
 package org.apache.curator.framework.recipes.cache;
 
 import com.google.common.annotations.VisibleForTesting;
-import com.google.common.base.Function;
 import com.google.common.base.Preconditions;
 import com.google.common.collect.ImmutableMap;
 import com.google.common.collect.Maps;
+import java.util.function.Supplier;
 import org.apache.curator.framework.CuratorFramework;
+import org.apache.curator.framework.WatcherRemoveCuratorFramework;
 import org.apache.curator.framework.api.BackgroundCallback;
+import org.apache.curator.framework.api.BackgroundPathable;
 import org.apache.curator.framework.api.CuratorEvent;
+import org.apache.curator.framework.api.Pathable;
 import org.apache.curator.framework.api.UnhandledErrorListener;
+import org.apache.curator.framework.api.Watchable;
 import org.apache.curator.framework.listen.Listenable;
-import org.apache.curator.framework.listen.ListenerContainer;
+import org.apache.curator.framework.listen.StandardListenerManager;
 import org.apache.curator.framework.state.ConnectionState;
 import org.apache.curator.framework.state.ConnectionStateListener;
-import org.apache.curator.utils.CloseableExecutorService;
+import org.apache.curator.utils.PathUtils;
 import org.apache.curator.utils.ThreadUtils;
 import org.apache.curator.utils.ZKPaths;
 import org.apache.zookeeper.KeeperException;
@@ -44,6 +48,8 @@ import org.slf4j.LoggerFactory;
 import java.io.Closeable;
 import java.util.ArrayList;
 import java.util.Collections;
+import java.util.Iterator;
+import java.util.LinkedList;
 import java.util.List;
 import java.util.Map;
 import java.util.concurrent.ConcurrentMap;
@@ -53,6 +59,7 @@ import java.util.concurrent.ThreadFactory;
 import java.util.concurrent.atomic.AtomicBoolean;
 import java.util.concurrent.atomic.AtomicLong;
 import java.util.concurrent.atomic.AtomicReference;
+import java.util.concurrent.atomic.AtomicReferenceFieldUpdater;
 
 import static com.google.common.base.Preconditions.checkNotNull;
 import static org.apache.curator.utils.PathUtils.validatePath;
@@ -65,10 +72,16 @@ import static org.apache.curator.utils.PathUtils.validatePath;
  * <p><b>IMPORTANT</b> - it's not possible to stay transactionally in sync. Users of this class must
  * be prepared for false-positives and false-negatives. Additionally, always use the version number
  * when updating data to avoid overwriting another process' change.</p>
+ *
+ * @deprecated replace by {@link org.apache.curator.framework.recipes.cache.CuratorCache}
  */
+@Deprecated
 public class TreeCache implements Closeable
 {
     private static final Logger LOG = LoggerFactory.getLogger(TreeCache.class);
+    private final boolean createParentNodes;
+    private final boolean disableZkWatches;
+    private final TreeCacheSelector selector;
 
     public static final class Builder
     {
@@ -76,8 +89,11 @@ public class TreeCache implements Closeable
         private final String path;
         private boolean cacheData = true;
         private boolean dataIsCompressed = false;
-        private CloseableExecutorService executorService = null;
+        private ExecutorService executorService = null;
         private int maxDepth = Integer.MAX_VALUE;
+        private boolean createParentNodes = false;
+        private boolean disableZkWatches = false;
+        private TreeCacheSelector selector = new DefaultTreeCacheSelector();
 
         private Builder(CuratorFramework client, String path)
         {
@@ -90,12 +106,12 @@ public class TreeCache implements Closeable
          */
         public TreeCache build()
         {
-            CloseableExecutorService executor = executorService;
+            ExecutorService executor = executorService;
             if ( executor == null )
             {
-                executor = new CloseableExecutorService(Executors.newSingleThreadExecutor(defaultThreadFactory));
+                executor = Executors.newSingleThreadExecutor(defaultThreadFactorySupplier.get());
             }
-            return new TreeCache(client, path, cacheData, dataIsCompressed, maxDepth, executor);
+            return new TreeCache(client, path, cacheData, dataIsCompressed, maxDepth, executor, createParentNodes, disableZkWatches, selector);
         }
 
         /**
@@ -121,28 +137,13 @@ public class TreeCache implements Closeable
          */
         public Builder setExecutor(ThreadFactory threadFactory)
         {
-            return setExecutor(new CloseableExecutorService(Executors.newSingleThreadExecutor(threadFactory)));
+            return setExecutor(Executors.newSingleThreadExecutor(threadFactory));
         }
 
         /**
          * Sets the executor to publish events; a default executor will be created if not specified.
          */
         public Builder setExecutor(ExecutorService executorService)
-        {
-            if ( executorService instanceof CloseableExecutorService )
-            {
-                return setExecutor((CloseableExecutorService)executorService);
-            }
-            else
-            {
-                return setExecutor(new CloseableExecutorService(executorService));
-            }
-        }
-
-        /**
-         * Sets the executor to publish events; a default executor will be created if not specified.
-         */
-        public Builder setExecutor(CloseableExecutorService executorService)
         {
             this.executorService = checkNotNull(executorService);
             return this;
@@ -159,6 +160,43 @@ public class TreeCache implements Closeable
             this.maxDepth = maxDepth;
             return this;
         }
+
+        /**
+         * By default, TreeCache does not auto-create parent nodes for the cached path. Change
+         * this behavior with this method. NOTE: parent nodes are created as containers
+         *
+         * @param createParentNodes true to create parent nodes
+         * @return this for chaining
+         */
+        public Builder setCreateParentNodes(boolean createParentNodes)
+        {
+            this.createParentNodes = createParentNodes;
+            return this;
+        }
+
+        /**
+         * By default, TreeCache creates {@link org.apache.zookeeper.ZooKeeper} watches for every created path.
+         * Change this behavior with this method.
+         * @param disableZkWatches true to disable zk watches
+         * @return this for chaining
+         */
+        public Builder disableZkWatches(boolean disableZkWatches)
+        {
+            this.disableZkWatches = disableZkWatches;
+            return this;
+        }
+
+        /**
+         * By default, {@link DefaultTreeCacheSelector} is used. Change the selector here.
+         *
+         * @param selector new selector
+         * @return this for chaining
+         */
+        public Builder setSelector(TreeCacheSelector selector)
+        {
+            this.selector = selector;
+            return this;
+        }
     }
 
     /**
@@ -179,19 +217,23 @@ public class TreeCache implements Closeable
         return new Builder(client, path);
     }
 
-    private enum NodeState
+    private static final ChildData DEAD = new ChildData("/", null, null);
+
+    static boolean isLive(ChildData cd)
     {
-        PENDING, LIVE, DEAD
+        return cd != null && cd != DEAD;
     }
 
-    private final class TreeNode implements Watcher, BackgroundCallback
+    private static final AtomicReferenceFieldUpdater<TreeNode, ChildData> childDataUpdater = AtomicReferenceFieldUpdater.newUpdater(TreeNode.class, ChildData.class, "childData");
+
+    private static final AtomicReferenceFieldUpdater<TreeNode, ConcurrentMap<String, TreeNode>> childrenUpdater = (AtomicReferenceFieldUpdater)AtomicReferenceFieldUpdater.newUpdater(TreeNode.class, ConcurrentMap.class, "children");
+
+    final class TreeNode implements Watcher, BackgroundCallback
     {
-        final AtomicReference<NodeState> nodeState = new AtomicReference<NodeState>(NodeState.PENDING);
+        volatile ChildData childData;
         final TreeNode parent;
         final String path;
-        final AtomicReference<Stat> stat = new AtomicReference<Stat>();
-        final AtomicReference<byte[]> data = new AtomicReference<byte[]>();
-        final AtomicReference<ConcurrentMap<String, TreeNode>> children = new AtomicReference<ConcurrentMap<String, TreeNode>>();
+        volatile ConcurrentMap<String, TreeNode> children;
         final int depth;
 
         TreeNode(String path, TreeNode parent)
@@ -203,7 +245,7 @@ public class TreeCache implements Closeable
 
         private void refresh() throws Exception
         {
-            if (depth < maxDepth)
+            if ((depth < maxDepth) && selector.traverseChildren(path))
             {
                 outstandingOps.addAndGet(2);
                 doRefreshData();
@@ -215,7 +257,7 @@ public class TreeCache implements Closeable
 
         private void refreshChildren() throws Exception
         {
-            if (depth < maxDepth)
+            if ((depth < maxDepth) && selector.traverseChildren(path))
             {
                 outstandingOps.incrementAndGet();
                 doRefreshChildren();
@@ -230,25 +272,40 @@ public class TreeCache implements Closeable
 
         private void doRefreshChildren() throws Exception
         {
-            client.getChildren().usingWatcher(this).inBackground(this).forPath(path);
+            if ( treeState.get() == TreeState.STARTED )
+            {
+                maybeWatch(client.getChildren()).forPath(path);
+            }
         }
 
         private void doRefreshData() throws Exception
         {
-            if ( dataIsCompressed )
+            if ( treeState.get() == TreeState.STARTED )
             {
-                client.getData().decompressed().usingWatcher(this).inBackground(this).forPath(path);
+                if ( dataIsCompressed )
+                {
+                    maybeWatch(client.getData().decompressed()).forPath(path);
+                }
+                else
+                {
+                    maybeWatch(client.getData()).forPath(path);
+                }
             }
-            else
-            {
-                client.getData().usingWatcher(this).inBackground(this).forPath(path);
+        }
+
+        private <T, P extends Watchable<BackgroundPathable<T>> & BackgroundPathable<T>> Pathable<T> maybeWatch(
+            P dataBuilder) {
+            if (disableZkWatches) {
+                return dataBuilder.inBackground(this);
+            } else {
+                return dataBuilder.usingWatcher(this).inBackground(this);
             }
         }
 
         void wasReconnected() throws Exception
         {
             refresh();
-            ConcurrentMap<String, TreeNode> childMap = children.get();
+            ConcurrentMap<String, TreeNode> childMap = children;
             if ( childMap != null )
             {
                 for ( TreeNode child : childMap.values() )
@@ -265,10 +322,12 @@ public class TreeCache implements Closeable
 
         void wasDeleted() throws Exception
         {
-            stat.set(null);
-            data.set(null);
-            client.clearWatcherReferences(this);
-            ConcurrentMap<String, TreeNode> childMap = children.getAndSet(null);
+            ChildData oldChildData = childDataUpdater.getAndSet(this, DEAD);
+            if ( oldChildData == DEAD )
+            {
+                return;
+            }
+            ConcurrentMap<String, TreeNode> childMap = childrenUpdater.getAndSet(this, null);
             if ( childMap != null )
             {
                 ArrayList<TreeNode> childCopy = new ArrayList<TreeNode>(childMap.values());
@@ -284,20 +343,20 @@ public class TreeCache implements Closeable
                 return;
             }
 
-            if ( nodeState.compareAndSet(NodeState.LIVE, NodeState.DEAD) )
+            if ( isLive(oldChildData) )
             {
-                publishEvent(TreeCacheEvent.Type.NODE_REMOVED, path);
+                publishEvent(TreeCacheEvent.Type.NODE_REMOVED, oldChildData, null);
             }
 
             if ( parent == null )
             {
                 // Root node; use an exist query to watch for existence.
-                client.checkExists().usingWatcher(this).inBackground().forPath(path);
+                maybeWatch(client.checkExists()).forPath(path);
             }
             else
             {
                 // Remove from parent if we're currently a child
-                ConcurrentMap<String, TreeNode> parentChildMap = parent.children.get();
+                ConcurrentMap<String, TreeNode> parentChildMap = parent.children;
                 if ( parentChildMap != null )
                 {
                     parentChildMap.remove(ZKPaths.getNodeFromPath(path), this);
@@ -308,6 +367,7 @@ public class TreeCache implements Closeable
         @Override
         public void process(WatchedEvent event)
         {
+            LOG.debug("process: {}", event);
             try
             {
                 switch ( event.getType() )
@@ -329,6 +389,7 @@ public class TreeCache implements Closeable
             }
             catch ( Exception e )
             {
+                ThreadUtils.checkInterrupted(e);
                 handleException(e);
             }
         }
@@ -336,6 +397,7 @@ public class TreeCache implements Closeable
         @Override
         public void processResult(CuratorFramework client, CuratorEvent event) throws Exception
         {
+            LOG.debug("processResult: {}", event);
             Stat newStat = event.getStat();
             switch ( event.getType() )
             {
@@ -343,23 +405,20 @@ public class TreeCache implements Closeable
                 Preconditions.checkState(parent == null, "unexpected EXISTS on non-root node");
                 if ( event.getResultCode() == KeeperException.Code.OK.intValue() )
                 {
-                    nodeState.compareAndSet(NodeState.DEAD, NodeState.PENDING);
+                    childDataUpdater.compareAndSet(this, DEAD, null);
                     wasCreated();
                 }
-                else if ( event.getResultCode() == KeeperException.Code.NONODE.intValue() )
-                {
-                    wasDeleted();
-                }
                 break;
             case CHILDREN:
                 if ( event.getResultCode() == KeeperException.Code.OK.intValue() )
                 {
-                    Stat oldStat = stat.get();
-                    if ( oldStat != null && oldStat.getMzxid() == newStat.getMzxid() )
+                    ChildData oldChildData = childData;
+                    //TODO consider doing update of cversion, pzxid, numChildren only
+                    if ( isLive(oldChildData) && oldChildData.getStat().getMzxid() == newStat.getMzxid() )
                     {
-                        // Only update stat if mzxid is different, otherwise we might obscure
+                        // Only update stat if mzxid is same, otherwise we might obscure
                         // GET_DATA event updates.
-                        stat.set(newStat);
+                        childDataUpdater.compareAndSet(this, oldChildData, new ChildData(oldChildData.getPath(), newStat, oldChildData.getData()));
                     }
 
                     if ( event.getChildren().isEmpty() )
@@ -367,13 +426,13 @@ public class TreeCache implements Closeable
                         break;
                     }
 
-                    ConcurrentMap<String, TreeNode> childMap = children.get();
-                    if ( childMap == null )
+                    ConcurrentMap<String, TreeNode> childMap = children;
+                    while ( childMap == null )
                     {
                         childMap = Maps.newConcurrentMap();
-                        if ( !children.compareAndSet(null, childMap) )
+                        if ( !childrenUpdater.compareAndSet(this, null, childMap) )
                         {
-                            childMap = children.get();
+                            childMap = children;
                         }
                     }
 
@@ -381,7 +440,7 @@ public class TreeCache implements Closeable
                     List<String> newChildren = new ArrayList<String>();
                     for ( String child : event.getChildren() )
                     {
-                        if ( !childMap.containsKey(child) )
+                        if ( !childMap.containsKey(child) && selector.acceptChild(ZKPaths.makePath(path, child)) )
                         {
                             newChildren.add(child);
                         }
@@ -406,21 +465,34 @@ public class TreeCache implements Closeable
             case GET_DATA:
                 if ( event.getResultCode() == KeeperException.Code.OK.intValue() )
                 {
-                    if ( cacheData )
+                    String eventPath = event.getPath();
+                    ChildData toPublish = new ChildData(eventPath, newStat, event.getData());
+                    ChildData toUpdate = cacheData ? toPublish : new ChildData(eventPath, newStat, null);
+                    while ( true )
                     {
-                        data.set(event.getData());
-                    }
-
-                    Stat oldStat = stat.getAndSet(newStat);
-                    if ( nodeState.compareAndSet(NodeState.PENDING, NodeState.LIVE) )
-                    {
-                        publishEvent(TreeCacheEvent.Type.NODE_ADDED, new ChildData(event.getPath(), newStat, event.getData()));
-                    }
-                    else
-                    {
-                        if ( oldStat == null || oldStat.getMzxid() != newStat.getMzxid() )
+                        final ChildData oldChildData = childData;
+                        // Ignore this event if we've already processed a newer update for this node.
+                        if ( isLive(oldChildData) && newStat.getMzxid() <= oldChildData.getStat().getMzxid() )
                         {
-                            publishEvent(TreeCacheEvent.Type.NODE_UPDATED, new ChildData(event.getPath(), newStat, event.getData()));
+                            break;
+                        }
+                        // Non-root nodes are not allowed to transition from dead -> live;
+                        // make sure this isn't a delayed response that came in after death.
+                        if ( parent != null && oldChildData == DEAD )
+                        {
+                            break;
+                        }
+                        if ( childDataUpdater.compareAndSet(this, oldChildData, toUpdate) )
+                        {
+                            if ( isLive(oldChildData) )
+                            {
+                                publishEvent(TreeCacheEvent.Type.NODE_UPDATED, toPublish, oldChildData);
+                            }
+                            else
+                            {
+                                publishEvent(TreeCacheEvent.Type.NODE_ADDED, toPublish, null);
+                            }
+                            break;
                         }
                     }
                 }
@@ -465,13 +537,13 @@ public class TreeCache implements Closeable
     private final AtomicBoolean isInitialized = new AtomicBoolean(false);
 
     private final TreeNode root;
-    private final CuratorFramework client;
-    private final CloseableExecutorService executorService;
+    private final WatcherRemoveCuratorFramework client;
+    private final ExecutorService executorService;
     private final boolean cacheData;
     private final boolean dataIsCompressed;
     private final int maxDepth;
-    private final ListenerContainer<TreeCacheListener> listeners = new ListenerContainer<TreeCacheListener>();
-    private final ListenerContainer<UnhandledErrorListener> errorListeners = new ListenerContainer<UnhandledErrorListener>();
+    private final StandardListenerManager<TreeCacheListener> listeners = StandardListenerManager.standard();
+    private final StandardListenerManager<UnhandledErrorListener> errorListeners = StandardListenerManager.standard();
     private final AtomicReference<TreeState> treeState = new AtomicReference<TreeState>(TreeState.LATENT);
 
     private final ConnectionStateListener connectionStateListener = new ConnectionStateListener()
@@ -483,7 +555,7 @@ public class TreeCache implements Closeable
         }
     };
 
-    static final ThreadFactory defaultThreadFactory = ThreadUtils.newThreadFactory("TreeCache");
+    private static final Supplier<ThreadFactory> defaultThreadFactorySupplier = () -> ThreadUtils.newThreadFactory("TreeCache");
 
     /**
      * Create a TreeCache for the given client and path with default options.
@@ -500,7 +572,7 @@ public class TreeCache implements Closeable
      */
     public TreeCache(CuratorFramework client, String path)
     {
-        this(client, path, true, false, Integer.MAX_VALUE, new CloseableExecutorService(Executors.newSingleThreadExecutor(defaultThreadFactory), true));
+        this(client, path, true, false, Integer.MAX_VALUE, Executors.newSingleThreadExecutor(defaultThreadFactorySupplier.get()), false, false, new DefaultTreeCacheSelector());
     }
 
     /**
@@ -509,15 +581,22 @@ public class TreeCache implements Closeable
      * @param cacheData        if true, node contents are cached in addition to the stat
      * @param dataIsCompressed if true, data in the path is compressed
      * @param executorService  Closeable ExecutorService to use for the TreeCache's background thread
+     * @param createParentNodes true to create parent nodes as containers
+     * @param disableZkWatches true to disable Zookeeper watches
+     * @param selector         the selector to use
      */
-    TreeCache(CuratorFramework client, String path, boolean cacheData, boolean dataIsCompressed, int maxDepth, final CloseableExecutorService executorService)
+    TreeCache(CuratorFramework client, String path, boolean cacheData, boolean dataIsCompressed, int maxDepth, final ExecutorService executorService, boolean createParentNodes, boolean disableZkWatches, TreeCacheSelector selector)
     {
+        this.createParentNodes = createParentNodes;
+        this.selector = Preconditions.checkNotNull(selector, "selector cannot be null");
         this.root = new TreeNode(validatePath(path), null);
-        this.client = client;
+        Preconditions.checkNotNull(client, "client cannot be null");
+        this.client = client.newWatcherRemoveCuratorFramework();
         this.cacheData = cacheData;
         this.dataIsCompressed = dataIsCompressed;
         this.maxDepth = maxDepth;
-        this.executorService = executorService;
+        this.disableZkWatches = disableZkWatches;
+        this.executorService = Preconditions.checkNotNull(executorService, "executorService cannot be null");
     }
 
     /**
@@ -529,6 +608,10 @@ public class TreeCache implements Closeable
     public TreeCache start() throws Exception
     {
         Preconditions.checkState(treeState.compareAndSet(TreeState.LATENT, TreeState.STARTED), "already started");
+        if ( createParentNodes )
+        {
+            client.createContainers(root.path);
+        }
         client.getConnectionStateListenable().addListener(connectionStateListener);
         if ( client.getZookeeperClient().isConnected() )
         {
@@ -545,15 +628,17 @@ public class TreeCache implements Closeable
     {
         if ( treeState.compareAndSet(TreeState.STARTED, TreeState.CLOSED) )
         {
+            client.removeWatchers();
             client.getConnectionStateListenable().removeListener(connectionStateListener);
             listeners.clear();
-            executorService.close();
+            executorService.shutdown();
             try
             {
                 root.wasDeleted();
             }
             catch ( Exception e )
             {
+                ThreadUtils.checkInterrupted(e);
                 handleException(e);
             }
         }
@@ -570,43 +655,46 @@ public class TreeCache implements Closeable
     }
 
     /**
-     * Allows catching unhandled errors in asynchornous operations.
+     * Allows catching unhandled errors in asynchronous operations.
      *
      * TODO: consider making public.
      */
     @VisibleForTesting
-    Listenable<UnhandledErrorListener> getUnhandledErrorListenable()
+    public Listenable<UnhandledErrorListener> getUnhandledErrorListenable()
     {
         return errorListeners;
     }
 
-    private TreeNode find(String fullPath)
+    private TreeNode find(String findPath)
     {
-        if ( !fullPath.startsWith(root.path) )
-        {
-            return null;
+        PathUtils.validatePath(findPath);
+        LinkedList<String> rootElements = new LinkedList<String>(ZKPaths.split(root.path));
+        LinkedList<String> findElements = new LinkedList<String>(ZKPaths.split(findPath));
+        while (!rootElements.isEmpty()) {
+            if (findElements.isEmpty()) {
+                // Target path shorter than root path
+                return null;
+            }
+            String nextRoot = rootElements.removeFirst();
+            String nextFind = findElements.removeFirst();
+            if (!nextFind.equals(nextRoot)) {
+                // Initial root path does not match
+                return null;
+            }
         }
 
         TreeNode current = root;
-        if ( fullPath.length() > root.path.length() )
-        {
-            if ( root.path.length() > 1 )
+        while (!findElements.isEmpty()) {
+            String nextFind = findElements.removeFirst();
+            ConcurrentMap<String, TreeNode> map = current.children;
+            if ( map == null )
             {
-                fullPath = fullPath.substring(root.path.length());
+                return null;
             }
-            List<String> split = ZKPaths.split(fullPath);
-            for ( String part : split )
+            current = map.get(nextFind);
+            if ( current == null )
             {
-                ConcurrentMap<String, TreeNode> map = current.children.get();
-                if ( map == null )
-                {
-                    return null;
-                }
-                current = map.get(part);
-                if ( current == null )
-                {
-                    return null;
-                }
+                return null;
             }
         }
         return current;
@@ -623,11 +711,11 @@ public class TreeCache implements Closeable
     public Map<String, ChildData> getCurrentChildren(String fullPath)
     {
         TreeNode node = find(fullPath);
-        if ( node == null || node.nodeState.get() != NodeState.LIVE )
+        if ( node == null || !isLive(node.childData) )
         {
             return null;
         }
-        ConcurrentMap<String, TreeNode> map = node.children.get();
+        ConcurrentMap<String, TreeNode> map = node.children;
         Map<String, ChildData> result;
         if ( map == null )
         {
@@ -638,10 +726,9 @@ public class TreeCache implements Closeable
             ImmutableMap.Builder<String, ChildData> builder = ImmutableMap.builder();
             for ( Map.Entry<String, TreeNode> entry : map.entrySet() )
             {
-                TreeNode childNode = entry.getValue();
-                ChildData childData = new ChildData(childNode.path, childNode.stat.get(), childNode.data.get());
-                // Double-check liveness after retreiving data.
-                if ( childNode.nodeState.get() == NodeState.LIVE )
+                ChildData childData = entry.getValue().childData;
+                // Double-check liveness after retrieving data.
+                if ( isLive(childData) )
                 {
                     builder.put(entry.getKey(), childData);
                 }
@@ -649,8 +736,8 @@ public class TreeCache implements Closeable
             result = builder.build();
         }
 
-        // Double-check liveness after retreiving children.
-        return node.nodeState.get() == NodeState.LIVE ? result : null;
+        // Double-check liveness after retrieving children.
+        return isLive(node.childData) ? result : null;
     }
 
     /**
@@ -664,31 +751,69 @@ public class TreeCache implements Closeable
     public ChildData getCurrentData(String fullPath)
     {
         TreeNode node = find(fullPath);
-        if ( node == null || node.nodeState.get() != NodeState.LIVE )
+        if ( node == null )
         {
             return null;
         }
-        ChildData result = new ChildData(node.path, node.stat.get(), node.data.get());
-        // Double-check liveness after retreiving data.
-        return node.nodeState.get() == NodeState.LIVE ? result : null;
+        ChildData result = node.childData;
+        return isLive(result) ? result : null;
+    }
+
+    /**
+     * Return an iterator over all nodes in the cache. There are no
+     * guarantees of accuracy; this is merely the most recent view of the data.
+     *
+     * @return a possibly-empty iterator of nodes in the cache
+     */
+    public Iterator<ChildData> iterator()
+    {
+        return new TreeCacheIterator(root);
+    }
+
+    /**
+     * Return the number of nodes in the cache. There are no
+     * guarantees of accuracy; this is merely the most recent view of the data.
+     *
+     * @return size
+     */
+    public int size()
+    {
+        return size(root);
+    }
+
+    private int size(TreeNode node)
+    {
+        int size;
+        if ( isLive(node.childData) )
+        {
+            size = 1;
+            if ( node.children != null )
+            {
+                for ( TreeNode child : node.children.values() )
+                {
+                    size += size(child);
+                }
+            }
+        }
+        else
+        {
+            size = 0;
+        }
+        return size;
     }
 
     private void callListeners(final TreeCacheEvent event)
     {
-        listeners.forEach(new Function<TreeCacheListener, Void>()
+        listeners.forEach(listener ->
         {
-            @Override
-            public Void apply(TreeCacheListener listener)
+            try
             {
-                try
-                {
-                    listener.childEvent(client, event);
-                }
-                catch ( Exception e )
-                {
-                    handleException(e);
-                }
-                return null;
+                listener.childEvent(client, event);
+            }
+            catch ( Exception e )
+            {
+                ThreadUtils.checkInterrupted(e);
+                handleException(e);
             }
         });
     }
@@ -704,20 +829,16 @@ public class TreeCache implements Closeable
         }
         else
         {
-            errorListeners.forEach(new Function<UnhandledErrorListener, Void>()
+            errorListeners.forEach(listener ->
             {
-                @Override
-                public Void apply(UnhandledErrorListener listener)
+                try
                 {
-                    try
-                    {
-                        listener.unhandledError("", e);
-                    }
-                    catch ( Exception e )
-                    {
-                        LOG.error("Exception handling exception", e);
-                    }
-                    return null;
+                    listener.unhandledError("", e);
+                }
+                catch ( Exception e2 )
+                {
+                    ThreadUtils.checkInterrupted(e2);
+                    LOG.error("Exception handling exception", e2);
                 }
             });
         }
@@ -732,6 +853,7 @@ public class TreeCache implements Closeable
             break;
 
         case LOST:
+            isInitialized.set(false);
             publishEvent(TreeCacheEvent.Type.CONNECTION_LOST);
             break;
 
@@ -742,6 +864,7 @@ public class TreeCache implements Closeable
             }
             catch ( Exception e )
             {
+                ThreadUtils.checkInterrupted(e);
                 handleException(e);
             }
             break;
@@ -754,6 +877,7 @@ public class TreeCache implements Closeable
             }
             catch ( Exception e )
             {
+                ThreadUtils.checkInterrupted(e);
                 handleException(e);
             }
             break;
@@ -765,34 +889,29 @@ public class TreeCache implements Closeable
         publishEvent(new TreeCacheEvent(type, null));
     }
 
-    private void publishEvent(TreeCacheEvent.Type type, String path)
+    private void publishEvent(TreeCacheEvent.Type type, ChildData data, ChildData oldData)
     {
-        publishEvent(new TreeCacheEvent(type, new ChildData(path, null, null)));
-    }
-
-    private void publishEvent(TreeCacheEvent.Type type, ChildData data)
-    {
-        publishEvent(new TreeCacheEvent(type, data));
+        publishEvent(new TreeCacheEvent(type, data, oldData));
     }
 
     private void publishEvent(final TreeCacheEvent event)
     {
         if ( treeState.get() != TreeState.CLOSED )
         {
+            LOG.debug("publishEvent: {}", event);
             executorService.submit(new Runnable()
             {
                 @Override
                 public void run()
                 {
+                    try
                     {
-                        try
-                        {
-                            callListeners(event);
-                        }
-                        catch ( Exception e )
-                        {
-                            handleException(e);
-                        }
+                        callListeners(event);
+                    }
+                    catch ( Exception e )
+                    {
+                        ThreadUtils.checkInterrupted(e);
+                        handleException(e);
                     }
                 }
             });
