@@ -34,6 +34,7 @@ import org.slf4j.LoggerFactory;
 import java.io.Closeable;
 import java.io.UnsupportedEncodingException;
 import java.util.Collection;
+import java.util.Iterator;
 import java.util.List;
 import java.util.concurrent.AbstractExecutorService;
 import java.util.concurrent.Callable;
@@ -45,6 +46,7 @@ import java.util.concurrent.Future;
 import java.util.concurrent.ThreadFactory;
 import java.util.concurrent.TimeUnit;
 import java.util.concurrent.atomic.AtomicBoolean;
+import java.util.concurrent.atomic.AtomicInteger;
 import java.util.concurrent.atomic.AtomicReference;
 import org.apache.curator.utils.PathUtils;
 
@@ -219,7 +221,7 @@ public class LeaderSelector implements Closeable
 
     /**
      * Re-queue an attempt for leadership. If this instance is already queued, nothing
-     * happens and false is returned. If the instance was not queued, it is re-qeued and true
+     * happens and false is returned. If the instance was not queued, it is re-queued and true
      * is returned
      *
      * @return true if re-queue is successful
@@ -230,7 +232,7 @@ public class LeaderSelector implements Closeable
         return internalRequeue();
     }
 
-    public synchronized boolean internalRequeue()
+    private synchronized boolean internalRequeue()
     {
         if ( !isQueued && (state.get() == State.STARTED) )
         {
@@ -302,17 +304,14 @@ public class LeaderSelector implements Closeable
         boolean isLeader = true;
         for ( String path : participantNodes )
         {
-            try
-            {
-                Participant participant = participantForPath(client, path, isLeader);
-                builder.add(participant);
-            }
-            catch ( KeeperException.NoNodeException ignore )
-            {
-                // ignore
-            }
+            Participant participant = participantForPath(client, path, isLeader);
 
-            isLeader = false;   // by definition the first node is the leader
+            if( participant != null )
+            {
+                builder.add(participant);
+
+                isLeader = false;   // by definition the first node is the leader
+            }
         }
 
         return builder.build();
@@ -341,11 +340,28 @@ public class LeaderSelector implements Closeable
 
     static Participant getLeader(CuratorFramework client, Collection<String> participantNodes) throws Exception
     {
+        Participant result = null;
+
         if ( participantNodes.size() > 0 )
         {
-            return participantForPath(client, participantNodes.iterator().next(), true);
+            Iterator<String> iter = participantNodes.iterator();
+            while ( iter.hasNext() )
+            {
+                result = participantForPath(client, iter.next(), true);
+
+                if ( result != null )
+                {
+                    break;
+                }
+            }
         }
-        return new Participant();
+
+        if( result == null )
+        {
+            result = new Participant();
+        }
+
+        return result;
     }
 
     /**
@@ -372,11 +388,21 @@ public class LeaderSelector implements Closeable
 
     private static Participant participantForPath(CuratorFramework client, String path, boolean markAsLeader) throws Exception
     {
-        byte[] bytes = client.getData().forPath(path);
-        String thisId = new String(bytes, "UTF-8");
-        return new Participant(thisId, markAsLeader);
+        try
+        {
+            byte[] bytes = client.getData().forPath(path);
+            String thisId = new String(bytes, "UTF-8");
+            return new Participant(thisId, markAsLeader);
+        }
+        catch ( KeeperException.NoNodeException e )
+        {
+            return null;
+        }
     }
 
+    @VisibleForTesting
+    volatile AtomicInteger failedMutexReleaseCount = null;
+
     @VisibleForTesting
     void doWork() throws Exception
     {
@@ -405,7 +431,7 @@ public class LeaderSelector implements Closeable
             }
             catch ( Throwable e )
             {
-                log.error("The leader threw an exception", e);
+                ThreadUtils.checkInterrupted(e);
             }
             finally
             {
@@ -417,20 +443,34 @@ public class LeaderSelector implements Closeable
             Thread.currentThread().interrupt();
             throw e;
         }
-        catch ( Exception e )
-        {
-            throw e;
-        }
         finally
         {
-            hasLeadership = false;
-            try
+            if ( hasLeadership )
             {
-                mutex.release();
-            }
-            catch ( Exception ignore )
-            {
-                // ignore errors - this is just a safety
+                hasLeadership = false;
+                boolean wasInterrupted = Thread.interrupted();  // clear any interrupted tatus so that mutex.release() works immediately
+                try
+                {
+                    mutex.release();
+                }
+                catch ( Exception e )
+                {
+                    if ( failedMutexReleaseCount != null )
+                    {
+                        failedMutexReleaseCount.incrementAndGet();
+                    }
+
+                    ThreadUtils.checkInterrupted(e);
+                    log.error("The leader threw an exception", e);
+                    // ignore errors - this is just a safety
+                }
+                finally
+                {
+                    if ( wasInterrupted )
+                    {
+                        Thread.currentThread().interrupt();
+                    }
+                }
             }
         }
     }
