@@ -19,18 +19,19 @@
 package org.apache.curator.framework.recipes.queue;
 
 import com.google.common.annotations.VisibleForTesting;
-import com.google.common.base.Function;
 import com.google.common.base.Preconditions;
 import com.google.common.collect.Lists;
-
-import org.apache.curator.utils.CloseableUtils;
 import org.apache.curator.framework.CuratorFramework;
 import org.apache.curator.framework.api.BackgroundCallback;
 import org.apache.curator.framework.api.CuratorEvent;
 import org.apache.curator.framework.api.CuratorEventType;
 import org.apache.curator.framework.imps.CuratorFrameworkState;
-import org.apache.curator.framework.listen.ListenerContainer;
+import org.apache.curator.framework.listen.Listenable;
+import org.apache.curator.framework.listen.StandardListenerManager;
 import org.apache.curator.framework.recipes.leader.LeaderSelector;
+import org.apache.curator.utils.CloseableUtils;
+import org.apache.curator.utils.PathUtils;
+import org.apache.curator.utils.ThreadUtils;
 import org.apache.curator.utils.ZKPaths;
 import org.apache.zookeeper.CreateMode;
 import org.apache.zookeeper.KeeperException;
@@ -49,7 +50,6 @@ import java.util.concurrent.ThreadFactory;
 import java.util.concurrent.TimeUnit;
 import java.util.concurrent.atomic.AtomicInteger;
 import java.util.concurrent.atomic.AtomicReference;
-import org.apache.curator.utils.PathUtils;
 
 /**
  * <p>An implementation of the Distributed Queue ZK recipe. Items put into the queue
@@ -80,7 +80,7 @@ public class DistributedQueue<T> implements QueueBase<T>
     private final boolean isProducerOnly;
     private final String lockPath;
     private final AtomicReference<ErrorMode> errorMode = new AtomicReference<ErrorMode>(ErrorMode.REQUEUE);
-    private final ListenerContainer<QueuePutListener<T>> putListenerContainer = new ListenerContainer<QueuePutListener<T>>();
+    private final StandardListenerManager<QueuePutListener<T>> putListenerContainer = StandardListenerManager.standard();
     private final AtomicInteger lastChildCount = new AtomicInteger(0);
     private final int maxItems;
     private final int finalFlushMs;
@@ -163,7 +163,7 @@ public class DistributedQueue<T> implements QueueBase<T>
 
         try
         {
-            client.create().creatingParentsIfNeeded().forPath(queuePath);
+            client.create().creatingParentContainersIfNeeded().forPath(queuePath);
         }
         catch ( KeeperException.NodeExistsException ignore )
         {
@@ -173,7 +173,7 @@ public class DistributedQueue<T> implements QueueBase<T>
         {
             try
             {
-                client.create().creatingParentsIfNeeded().forPath(lockPath);
+                client.create().creatingParentContainersIfNeeded().forPath(lockPath);
             }
             catch ( KeeperException.NodeExistsException ignore )
             {
@@ -232,7 +232,7 @@ public class DistributedQueue<T> implements QueueBase<T>
      * @return put listener container
      */
     @Override
-    public ListenerContainer<QueuePutListener<T>> getPutListenerContainer()
+    public Listenable<QueuePutListener<T>> getPutListenerContainer()
     {
         return putListenerContainer;
     }
@@ -406,25 +406,16 @@ public class DistributedQueue<T> implements QueueBase<T>
             putCount.decrementAndGet();
             putCount.notifyAll();
         }
-        putListenerContainer.forEach
-        (
-            new Function<QueuePutListener<T>, Void>()
+        putListenerContainer.forEach(listener -> {
+            if ( item != null )
             {
-                @Override
-                public Void apply(QueuePutListener<T> listener)
-                {
-                    if ( item != null )
-                    {
-                        listener.putCompleted(item);
-                    }
-                    else
-                    {
-                        listener.putMultiCompleted(givenMultiItem);
-                    }
-                    return null;
-                }
+                listener.putCompleted(item);
             }
-        );
+            else
+            {
+                listener.putMultiCompleted(givenMultiItem);
+            }
+        });
     }
 
     private void doPutInBackground(final T item, String path, final MultiItem<T> givenMultiItem, byte[] bytes) throws Exception
@@ -448,25 +439,16 @@ public class DistributedQueue<T> implements QueueBase<T>
                     }
                 }
 
-                putListenerContainer.forEach
-                (
-                    new Function<QueuePutListener<T>, Void>()
+                putListenerContainer.forEach(listener -> {
+                    if ( item != null )
                     {
-                        @Override
-                        public Void apply(QueuePutListener<T> listener)
-                        {
-                            if ( item != null )
-                            {
-                                listener.putCompleted(item);
-                            }
-                            else
-                            {
-                                listener.putMultiCompleted(givenMultiItem);
-                            }
-                            return null;
-                        }
+                        listener.putCompleted(item);
                     }
-                );
+                    else
+                    {
+                        listener.putMultiCompleted(givenMultiItem);
+                    }
+                });
             }
         };
         internalCreateNode(path, bytes, callback);
@@ -544,34 +526,39 @@ public class DistributedQueue<T> implements QueueBase<T>
         long         maxWaitMs = -1;
         try
         {
-            while ( !Thread.currentThread().isInterrupted()  )
+            while ( state.get() == State.STARTED  )
             {
-                ChildrenCache.Data      data = (maxWaitMs > 0) ? childrenCache.blockingNextGetData(currentVersion, maxWaitMs, TimeUnit.MILLISECONDS) : childrenCache.blockingNextGetData(currentVersion);
-                currentVersion = data.version;
-
-                List<String>        children = Lists.newArrayList(data.children);
-                sortChildren(children); // makes sure items are processed in the correct order
-
-                if ( children.size() > 0 )
+                try
                 {
-                    maxWaitMs = getDelay(children.get(0));
-                    if ( maxWaitMs > 0 )
+                    ChildrenCache.Data      data = (maxWaitMs > 0) ? childrenCache.blockingNextGetData(currentVersion, maxWaitMs, TimeUnit.MILLISECONDS) : childrenCache.blockingNextGetData(currentVersion);
+                    currentVersion = data.version;
+
+                    List<String>        children = Lists.newArrayList(data.children);
+                    sortChildren(children); // makes sure items are processed in the correct order
+
+                    if ( children.size() > 0 )
+                    {
+                        maxWaitMs = getDelay(children.get(0));
+                        if ( maxWaitMs > 0 )
+                        {
+                            continue;
+                        }
+                    }
+                    else
                     {
                         continue;
                     }
-                }
-                else
-                {
-                    continue;
-                }
 
-                processChildren(children, currentVersion);
+                    processChildren(children, currentVersion);
+                }
+                catch ( InterruptedException e )
+                {
+                    // swallow the interrupt as it's only possible from either a background
+                    // operation and, thus, doesn't apply to this loop or the instance
+                    // is being closed in which case the while test will get it
+                }
             }
         }
-        catch ( InterruptedException ignore )
-        {
-            Thread.currentThread().interrupt();
-        }
         catch ( Exception e )
         {
             log.error("Exception caught in background handler", e);
@@ -633,6 +620,7 @@ public class DistributedQueue<T> implements QueueBase<T>
                         }
                         catch ( Exception e )
                         {
+                            ThreadUtils.checkInterrupted(e);
                             log.error("Error processing message at " + itemNode, e);
                         }
                         finally
@@ -663,6 +651,7 @@ public class DistributedQueue<T> implements QueueBase<T>
         }
         catch ( Throwable e )
         {
+            ThreadUtils.checkInterrupted(e);
             log.error("Corrupted queue item: " + itemNode, e);
             return resultCode;
         }
@@ -681,6 +670,7 @@ public class DistributedQueue<T> implements QueueBase<T>
             }
             catch ( Throwable e )
             {
+                ThreadUtils.checkInterrupted(e);
                 log.error("Exception processing queue item: " + itemNode, e);
                 if ( errorMode.get() == ErrorMode.REQUEUE )
                 {
@@ -756,7 +746,7 @@ public class DistributedQueue<T> implements QueueBase<T>
                 client.inTransaction()
                     .delete().forPath(itemPath)
                     .and()
-                    .create().withMode(CreateMode.PERSISTENT_SEQUENTIAL).forPath(makeItemPath(), bytes)
+                    .create().withMode(CreateMode.PERSISTENT_SEQUENTIAL).forPath(makeRequeueItemPath(itemPath), bytes)
                     .and()
                     .commit();
             }
@@ -789,4 +779,9 @@ public class DistributedQueue<T> implements QueueBase<T>
 
         return false;
     }
+
+    protected String makeRequeueItemPath(String itemPath)
+    {
+        return makeItemPath();
+    }
 }
