@@ -19,21 +19,33 @@
 
 package org.apache.curator.framework.recipes.leader;
 
+import static org.junit.jupiter.api.Assertions.assertEquals;
+import static org.junit.jupiter.api.Assertions.assertFalse;
+import static org.junit.jupiter.api.Assertions.assertNotNull;
+import static org.junit.jupiter.api.Assertions.assertNotSame;
+import static org.junit.jupiter.api.Assertions.assertTrue;
+import static org.junit.jupiter.api.Assertions.fail;
 import com.google.common.collect.Lists;
+import com.google.common.collect.Queues;
+import com.google.common.collect.Sets;
 import org.apache.curator.framework.CuratorFramework;
 import org.apache.curator.framework.CuratorFrameworkFactory;
 import org.apache.curator.framework.state.ConnectionState;
+import org.apache.curator.framework.state.ConnectionStateListener;
+import org.apache.curator.framework.state.SessionConnectionStateErrorPolicy;
+import org.apache.curator.framework.state.StandardConnectionStateErrorPolicy;
 import org.apache.curator.retry.RetryOneTime;
 import org.apache.curator.test.BaseClassForTests;
-import org.apache.curator.test.KillSession;
 import org.apache.curator.test.TestingServer;
 import org.apache.curator.test.Timing;
+import org.apache.curator.test.compatibility.Timing2;
 import org.apache.curator.utils.CloseableUtils;
-import org.testng.Assert;
-import org.testng.annotations.Test;
-import org.testng.internal.annotations.Sets;
+import org.junit.jupiter.api.Test;
+
+import java.util.Arrays;
 import java.util.List;
 import java.util.Set;
+import java.util.concurrent.ArrayBlockingQueue;
 import java.util.concurrent.BlockingQueue;
 import java.util.concurrent.CountDownLatch;
 import java.util.concurrent.LinkedBlockingQueue;
@@ -41,12 +53,210 @@ import java.util.concurrent.Semaphore;
 import java.util.concurrent.TimeUnit;
 import java.util.concurrent.atomic.AtomicInteger;
 
-import static org.testng.Assert.fail;
-
 public class TestLeaderSelector extends BaseClassForTests
 {
     private static final String PATH_NAME = "/one/two/me";
 
+    @Test
+    public void testInterruption() throws Exception
+    {
+        Timing2 timing = new Timing2();
+        LeaderSelector selector = null;
+        CuratorFramework client = CuratorFrameworkFactory.newClient(server.getConnectString(), new RetryOneTime(1));
+        try
+        {
+            client.start();
+
+            CountDownLatch exitLatch = new CountDownLatch(1);
+            BlockingQueue<Thread> threadExchange = new ArrayBlockingQueue<>(1);
+            LeaderSelectorListener listener = new LeaderSelectorListenerAdapter()
+            {
+                @Override
+                public void takeLeadership(CuratorFramework client) throws Exception
+                {
+                    threadExchange.put(Thread.currentThread());
+                    try
+                    {
+                        Thread.currentThread().join();
+                    }
+                    finally
+                    {
+                        exitLatch.countDown();
+                    }
+                }
+            };
+            selector = new LeaderSelector(client, PATH_NAME, listener);
+            selector.failedMutexReleaseCount = new AtomicInteger();
+            selector.start();
+            Thread leaderThread = timing.takeFromQueue(threadExchange);
+            leaderThread.interrupt();
+            assertTrue(timing.awaitLatch(exitLatch));
+            timing.sleepABit(); // wait for leader selector to clear nodes
+            assertEquals(0, selector.failedMutexReleaseCount.get());
+        }
+        finally
+        {
+            CloseableUtils.closeQuietly(selector);
+            CloseableUtils.closeQuietly(client);
+        }
+    }
+
+    @Test
+    public void testErrorPolicies() throws Exception
+    {
+        Timing2 timing = new Timing2();
+        LeaderSelector selector = null;
+        CuratorFramework client = CuratorFrameworkFactory
+            .builder()
+            .connectString(server.getConnectString())
+            .connectionTimeoutMs(timing.connection())
+            .sessionTimeoutMs(timing.session())
+            .retryPolicy(new RetryOneTime(1))
+            .connectionStateErrorPolicy(new StandardConnectionStateErrorPolicy())
+            .build();
+        try
+        {
+            final BlockingQueue<String> changes = Queues.newLinkedBlockingQueue();
+
+            ConnectionStateListener stateListener = new ConnectionStateListener()
+            {
+                @Override
+                public void stateChanged(CuratorFramework client, ConnectionState newState)
+                {
+                    changes.add(newState.name());
+                }
+            };
+            client.getConnectionStateListenable().addListener(stateListener);
+            client.start();
+            LeaderSelectorListener listener = new LeaderSelectorListenerAdapter()
+            {
+                @Override
+                public void takeLeadership(CuratorFramework client) throws Exception
+                {
+                    changes.add("leader");
+                    try
+                    {
+                        Thread.currentThread().join();
+                    }
+                    catch ( InterruptedException e )
+                    {
+                        changes.add("release");
+                        Thread.currentThread().interrupt();
+                    }
+                }
+            };
+            selector = new LeaderSelector(client, "/test", listener);
+            selector.start();
+
+            assertEquals(changes.poll(timing.forWaiting().milliseconds(), TimeUnit.MILLISECONDS), ConnectionState.CONNECTED.name());
+            assertEquals(changes.poll(timing.forWaiting().milliseconds(), TimeUnit.MILLISECONDS), "leader");
+            server.close();
+            List<String> next = Lists.newArrayList();
+            next.add(changes.poll(timing.forSessionSleep().milliseconds(), TimeUnit.MILLISECONDS));
+            next.add(changes.poll(timing.forSessionSleep().milliseconds(), TimeUnit.MILLISECONDS));
+            assertTrue(next.equals(Arrays.asList(ConnectionState.SUSPENDED.name(), "release")) || next.equals(Arrays.asList("release", ConnectionState.SUSPENDED.name())), next.toString());
+            assertEquals(changes.poll(timing.forSessionSleep().milliseconds(), TimeUnit.MILLISECONDS), ConnectionState.LOST.name());
+
+            selector.close();
+            client.close();
+            timing.sleepABit();
+            changes.clear();
+
+            server = new TestingServer();
+            client = CuratorFrameworkFactory
+                .builder()
+                .connectString(server.getConnectString())
+                .connectionTimeoutMs(timing.connection())
+                .sessionTimeoutMs(timing.session())
+                .retryPolicy(new RetryOneTime(1))
+                .connectionStateErrorPolicy(new SessionConnectionStateErrorPolicy())
+                .build();
+            client.getConnectionStateListenable().addListener(stateListener);
+            client.start();
+            selector = new LeaderSelector(client, "/test", listener);
+            selector.start();
+
+            assertEquals(changes.poll(timing.forWaiting().milliseconds(), TimeUnit.MILLISECONDS), ConnectionState.CONNECTED.name());
+            assertEquals(changes.poll(timing.forWaiting().milliseconds(), TimeUnit.MILLISECONDS), "leader");
+            server.stop();
+            assertEquals(changes.poll(timing.forWaiting().milliseconds(), TimeUnit.MILLISECONDS), ConnectionState.SUSPENDED.name());
+            next = Lists.newArrayList();
+            next.add(changes.poll(timing.forSessionSleep().milliseconds(), TimeUnit.MILLISECONDS));
+            next.add(changes.poll(timing.forSessionSleep().milliseconds(), TimeUnit.MILLISECONDS));
+            assertTrue(next.equals(Arrays.asList(ConnectionState.LOST.name(), "release")) || next.equals(Arrays.asList("release", ConnectionState.LOST.name())), next.toString());
+        }
+        finally
+        {
+            CloseableUtils.closeQuietly(selector);
+            CloseableUtils.closeQuietly(client);
+        }
+    }
+
+    @Test
+    public void testLeaderNodeDeleteOnInterrupt() throws Exception
+    {
+        Timing2 timing = new Timing2();
+        LeaderSelector selector = null;
+        CuratorFramework client = null;
+        try
+        {
+            client = CuratorFrameworkFactory.newClient(server.getConnectString(), timing.session(), timing.connection(), new RetryOneTime(1));
+            final CountDownLatch reconnectedLatch = new CountDownLatch(1);
+            ConnectionStateListener connectionStateListener = new ConnectionStateListener()
+            {
+                @Override
+                public void stateChanged(CuratorFramework client, ConnectionState newState)
+                {
+                    if ( newState == ConnectionState.RECONNECTED )
+                    {
+                        reconnectedLatch.countDown();
+                    }
+                }
+            };
+            client.getConnectionStateListenable().addListener(connectionStateListener);
+            client.start();
+
+            final BlockingQueue<Thread> queue = new ArrayBlockingQueue<Thread>(1);
+            LeaderSelectorListener listener = new LeaderSelectorListener()
+            {
+                @Override
+                public void takeLeadership(CuratorFramework client) throws Exception
+                {
+                    queue.add(Thread.currentThread());
+                    try
+                    {
+                        Thread.currentThread().join();
+                    }
+                    catch ( InterruptedException e )
+                    {
+                        Thread.currentThread().interrupt();
+                    }
+                }
+
+                @Override
+                public void stateChanged(CuratorFramework client, ConnectionState newState)
+                {
+                }
+            };
+            selector = new LeaderSelector(client, "/leader", listener);
+            selector.start();
+
+            Thread leaderThread = timing.takeFromQueue(queue);
+            server.stop();
+            leaderThread.interrupt();
+            server.restart();
+            assertTrue(timing.awaitLatch(reconnectedLatch));
+            timing.sleepABit();
+
+            assertEquals(client.getChildren().forPath("/leader").size(), 0);
+        }
+        finally
+        {
+            CloseableUtils.closeQuietly(selector);
+            CloseableUtils.closeQuietly(client);
+        }
+    }
+
     @Test
     public void testInterruptLeadershipWithRequeue() throws Exception
     {
@@ -72,10 +282,10 @@ public class TestLeaderSelector extends BaseClassForTests
             selector.autoRequeue();
             selector.start();
 
-            Assert.assertTrue(timing.acquireSemaphore(semaphore));
+            assertTrue(timing.acquireSemaphore(semaphore));
             selector.interruptLeadership();
 
-            Assert.assertTrue(timing.acquireSemaphore(semaphore));
+            assertTrue(timing.acquireSemaphore(semaphore));
         }
         finally
         {
@@ -121,9 +331,9 @@ public class TestLeaderSelector extends BaseClassForTests
             selector = new LeaderSelector(client, "/leader", listener);
             selector.start();
 
-            Assert.assertTrue(timing.awaitLatch(isLeaderLatch));
+            assertTrue(timing.awaitLatch(isLeaderLatch));
             selector.interruptLeadership();
-            Assert.assertTrue(timing.awaitLatch(losingLeaderLatch));
+            assertTrue(timing.awaitLatch(losingLeaderLatch));
         }
         finally
         {
@@ -176,16 +386,16 @@ public class TestLeaderSelector extends BaseClassForTests
 
             selector.start();
 
-            Assert.assertTrue(timing.awaitLatch(debugLeadershipLatch));
+            assertTrue(timing.awaitLatch(debugLeadershipLatch));
             server.stop();
-            Assert.assertTrue(timing.awaitLatch(lostLatch));
+            assertTrue(timing.awaitLatch(lostLatch));
             timing.sleepABit();
             debugLeadershipWaitLatch.countDown();
 
             server.restart();
-            Assert.assertTrue(timing.awaitLatch(reconnectedLatch));
+            assertTrue(timing.awaitLatch(reconnectedLatch));
 
-            Assert.assertFalse(takeLeadershipLatch.await(3, TimeUnit.SECONDS));
+            assertFalse(takeLeadershipLatch.await(3, TimeUnit.SECONDS));
         }
         finally
         {
@@ -223,7 +433,7 @@ public class TestLeaderSelector extends BaseClassForTests
             selector.autoRequeue();
             selector.start();
 
-            Assert.assertTrue(timing.acquireSemaphore(semaphore, 2));
+            assertTrue(timing.acquireSemaphore(semaphore, 2));
         }
         finally
         {
@@ -322,20 +532,20 @@ public class TestLeaderSelector extends BaseClassForTests
             leaderSelector1.start();
             leaderSelector2.start();
 
-            Assert.assertTrue(timing.acquireSemaphore(semaphore, 1));
+            assertTrue(timing.acquireSemaphore(semaphore, 1));
 
-            KillSession.kill(client.getZookeeperClient().getZooKeeper(), server.getConnectString());
+            client.getZookeeperClient().getZooKeeper().getTestable().injectSessionExpiration();
 
-            Assert.assertTrue(timing.awaitLatch(interruptedLatch));
+            assertTrue(timing.awaitLatch(interruptedLatch));
             timing.sleepABit();
 
             boolean requeued1 = leaderSelector1.requeue();
             boolean requeued2 = leaderSelector2.requeue();
-            Assert.assertTrue(requeued1);
-            Assert.assertTrue(requeued2);
+            assertTrue(requeued1);
+            assertTrue(requeued2);
 
-            Assert.assertTrue(timing.acquireSemaphore(semaphore, 1));
-            Assert.assertEquals(leaderCount.get(), 1);
+            assertTrue(timing.acquireSemaphore(semaphore, 1));
+            assertEquals(leaderCount.get(), 1);
 
             if ( leaderSelector1.hasLeadership() )
             {
@@ -353,8 +563,8 @@ public class TestLeaderSelector extends BaseClassForTests
             }
 
             // Verify that the other leader took over leadership.
-            Assert.assertTrue(timing.acquireSemaphore(semaphore, 1));
-            Assert.assertEquals(leaderCount.get(), 1);
+            assertTrue(timing.acquireSemaphore(semaphore, 1));
+            assertEquals(leaderCount.get(), 1);
 
             if ( !leaderSelector1Closed )
             {
@@ -426,17 +636,17 @@ public class TestLeaderSelector extends BaseClassForTests
             leaderSelector1.start();
             leaderSelector2.start();
 
-            Assert.assertTrue(timing.acquireSemaphore(semaphore, 1));
+            assertTrue(timing.acquireSemaphore(semaphore, 1));
 
             int port = server.getPort();
             server.stop();
             timing.sleepABit();
             server = new TestingServer(port);
-            Assert.assertTrue(timing.awaitLatch(interruptedLatch));
+            assertTrue(timing.awaitLatch(interruptedLatch));
             timing.sleepABit();
 
-            Assert.assertTrue(timing.acquireSemaphore(semaphore, 1));
-            Assert.assertEquals(leaderCount.get(), 1);
+            assertTrue(timing.acquireSemaphore(semaphore, 1));
+            assertEquals(leaderCount.get(), 1);
 
             if ( leaderSelector1.hasLeadership() )
             {
@@ -454,8 +664,8 @@ public class TestLeaderSelector extends BaseClassForTests
             }
 
             // Verify that the other leader took over leadership.
-            Assert.assertTrue(timing.acquireSemaphore(semaphore, 1));
-            Assert.assertEquals(leaderCount.get(), 1);
+            assertTrue(timing.acquireSemaphore(semaphore, 1));
+            assertEquals(leaderCount.get(), 1);
 
             if ( !leaderSelector1Closed )
             {
@@ -516,7 +726,7 @@ public class TestLeaderSelector extends BaseClassForTests
                 Thread.sleep(1000);
             }
 
-            Assert.assertNotSame(leaderSelector1.hasLeadership(), leaderSelector2.hasLeadership());
+            assertNotSame(leaderSelector1.hasLeadership(), leaderSelector2.hasLeadership());
 
             LeaderSelector positiveLeader;
             LeaderSelector negativeLeader;
@@ -533,12 +743,12 @@ public class TestLeaderSelector extends BaseClassForTests
 
             negativeLeader.close();
             Thread.sleep(1000);
-            Assert.assertNotSame(positiveLeader.hasLeadership(), negativeLeader.hasLeadership());
-            Assert.assertTrue(positiveLeader.hasLeadership());
+            assertNotSame(positiveLeader.hasLeadership(), negativeLeader.hasLeadership());
+            assertTrue(positiveLeader.hasLeadership());
 
             positiveLeader.close();
             Thread.sleep(1000);
-            Assert.assertFalse(positiveLeader.hasLeadership());
+            assertFalse(positiveLeader.hasLeadership());
         }
         finally
         {
@@ -598,7 +808,7 @@ public class TestLeaderSelector extends BaseClassForTests
                 while ( localLeaderList.size() != (i * selectors.size()) )
                 {
                     Integer polledIndex = leaderList.poll(10, TimeUnit.SECONDS);
-                    Assert.assertNotNull(polledIndex);
+                    assertNotNull(polledIndex);
                     localLeaderList.add(polledIndex);
                 }
                 timing.sleepABit();
@@ -615,10 +825,10 @@ public class TestLeaderSelector extends BaseClassForTests
                 Set<Integer> uniques = Sets.newHashSet();
                 for ( int j = 0; j < selectors.size(); ++j )
                 {
-                    Assert.assertTrue(localLeaderList.size() > 0);
+                    assertTrue(localLeaderList.size() > 0);
 
                     int thisIndex = localLeaderList.remove(0);
-                    Assert.assertFalse(uniques.contains(thisIndex));
+                    assertFalse(uniques.contains(thisIndex));
                     uniques.add(thisIndex);
                 }
             }
