@@ -19,20 +19,26 @@
 
 package org.apache.curator.framework.recipes.locks;
 
+import static org.junit.jupiter.api.Assertions.assertEquals;
+import static org.junit.jupiter.api.Assertions.assertFalse;
+import static org.junit.jupiter.api.Assertions.assertNull;
+import static org.junit.jupiter.api.Assertions.assertTrue;
+import static org.junit.jupiter.api.Assertions.fail;
 import com.google.common.collect.Lists;
-import org.apache.curator.test.BaseClassForTests;
-import org.apache.curator.utils.CloseableUtils;
 import org.apache.curator.framework.CuratorFramework;
 import org.apache.curator.framework.CuratorFrameworkFactory;
 import org.apache.curator.framework.state.ConnectionState;
 import org.apache.curator.framework.state.ConnectionStateListener;
 import org.apache.curator.retry.ExponentialBackoffRetry;
-import org.apache.curator.retry.RetryOneTime;
-import org.apache.curator.test.KillSession;
+import org.apache.curator.test.BaseClassForTests;
 import org.apache.curator.test.TestingServer;
 import org.apache.curator.test.Timing;
-import org.testng.Assert;
-import org.testng.annotations.Test;
+import org.apache.curator.test.compatibility.Timing2;
+import org.apache.curator.utils.CloseableUtils;
+import org.apache.curator.utils.ZKPaths;
+import org.apache.zookeeper.KeeperException;
+import org.junit.jupiter.api.Test;
+
 import java.util.List;
 import java.util.concurrent.Callable;
 import java.util.concurrent.CountDownLatch;
@@ -47,11 +53,35 @@ import java.util.concurrent.atomic.AtomicReference;
 
 public abstract class TestInterProcessMutexBase extends BaseClassForTests
 {
+    protected static final String LOCK_BASE_PATH = "/locks";
+
     private volatile CountDownLatch waitLatchForBar = null;
     private volatile CountDownLatch countLatchForBar = null;
 
     protected abstract InterProcessLock makeLock(CuratorFramework client);
 
+    @Test
+    public void testLocker() throws Exception
+    {
+        final Timing timing = new Timing();
+        final CuratorFramework client = CuratorFrameworkFactory.newClient(server.getConnectString(), timing.session(), timing.connection(), new ExponentialBackoffRetry(100, 3));
+        try
+        {
+            client.start();
+
+            InterProcessLock lock = makeLock(client);
+            try ( Locker locker = new Locker(lock, timing.milliseconds(), TimeUnit.MILLISECONDS) )
+            {
+                assertTrue(lock.isAcquiredInThisProcess());
+            }
+            assertFalse(lock.isAcquiredInThisProcess());
+        }
+        finally
+        {
+            CloseableUtils.closeQuietly(client);
+        }
+    }
+
     @Test
     public void testWaitingProcessKilledServer() throws Exception
     {
@@ -67,7 +97,7 @@ public abstract class TestInterProcessMutexBase extends BaseClassForTests
                 @Override
                 public void stateChanged(CuratorFramework client, ConnectionState newState)
                 {
-                    if ( newState == ConnectionState.LOST )
+                    if ( !newState.isConnected() )
                     {
                         latch.countDown();
                     }
@@ -76,6 +106,7 @@ public abstract class TestInterProcessMutexBase extends BaseClassForTests
             client.getConnectionStateListenable().addListener(listener);
 
             final AtomicBoolean isFirst = new AtomicBoolean(true);
+            final Object result = new Object();
             ExecutorCompletionService<Object> service = new ExecutorCompletionService<Object>(Executors.newFixedThreadPool(2));
             for ( int i = 0; i < 2; ++i )
             {
@@ -95,7 +126,7 @@ public abstract class TestInterProcessMutexBase extends BaseClassForTests
                                     timing.sleepABit();
 
                                     server.stop();
-                                    Assert.assertTrue(timing.awaitLatch(latch));
+                                    assertTrue(timing.forWaiting().awaitLatch(latch));
                                     server.restart();
                                 }
                             }
@@ -105,12 +136,12 @@ public abstract class TestInterProcessMutexBase extends BaseClassForTests
                                 {
                                     lock.release();
                                 }
-                                catch ( Exception e )
+                                catch ( KeeperException.SessionExpiredException dummy )
                                 {
-                                    // ignore
+                                    // happens sometimes with a few tests - ignore
                                 }
                             }
-                            return null;
+                            return result;
                         }
                     }
                 );
@@ -118,7 +149,7 @@ public abstract class TestInterProcessMutexBase extends BaseClassForTests
 
             for ( int i = 0; i < 2; ++i )
             {
-                service.take().get(timing.forWaiting().milliseconds(), TimeUnit.MILLISECONDS);
+                assertEquals(service.take().get(timing.forWaiting().milliseconds(), TimeUnit.MILLISECONDS), result);
             }
         }
         finally
@@ -130,7 +161,7 @@ public abstract class TestInterProcessMutexBase extends BaseClassForTests
     @Test
     public void testKilledSession() throws Exception
     {
-        final Timing timing = new Timing();
+        final Timing2 timing = new Timing2();
 
         CuratorFramework client = CuratorFrameworkFactory.newClient(server.getConnectString(), timing.session(), timing.connection(), new ExponentialBackoffRetry(100, 3));
         client.start();
@@ -171,13 +202,89 @@ public abstract class TestInterProcessMutexBase extends BaseClassForTests
                     }
                 );
 
-            Assert.assertTrue(timing.acquireSemaphore(semaphore, 1));
-            KillSession.kill(client.getZookeeperClient().getZooKeeper(), server.getConnectString());
-            Assert.assertTrue(timing.acquireSemaphore(semaphore, 1));
+            assertTrue(timing.acquireSemaphore(semaphore, 1));
+            client.getZookeeperClient().getZooKeeper().getTestable().injectSessionExpiration();
+            assertTrue(timing.forSessionSleep().acquireSemaphore(semaphore, 1));
         }
         finally
         {
-            client.close();
+            CloseableUtils.closeQuietly(client);
+        }
+    }
+
+    @Test
+    public void testContainerCleanup() throws Exception
+    {
+        if ( !ZKPaths.hasContainerSupport() )
+        {
+            System.out.println("ZooKeeper version does not support Containers. Skipping test");
+            return;
+        }
+
+        server.close();
+
+        System.setProperty("znode.container.checkIntervalMs", "10");
+        try
+        {
+            server = new TestingServer();
+
+            final int THREAD_QTY = 10;
+
+            ExecutorService service = null;
+            final CuratorFramework client = CuratorFrameworkFactory.newClient(server.getConnectString(), new ExponentialBackoffRetry(100, 3));
+            try
+            {
+                client.start();
+
+                List<Future<Object>> threads = Lists.newArrayList();
+                service = Executors.newCachedThreadPool();
+                for ( int i = 0; i < THREAD_QTY; ++i )
+                {
+                    Future<Object> t = service.submit
+                    (
+                        new Callable<Object>()
+                        {
+                            @Override
+                            public Object call() throws Exception
+                            {
+                                InterProcessLock lock = makeLock(client);
+                                lock.acquire();
+                                try
+                                {
+                                    Thread.sleep(10);
+                                }
+                                finally
+                                {
+                                    lock.release();
+                                }
+                                return null;
+                            }
+                        }
+                    );
+                    threads.add(t);
+                }
+
+                for ( Future<Object> t : threads )
+                {
+                    t.get();
+                }
+
+                new Timing().sleepABit();
+
+                assertNull(client.checkExists().forPath(LOCK_BASE_PATH));
+            }
+            finally
+            {
+                if ( service != null )
+                {
+                    service.shutdownNow();
+                }
+                CloseableUtils.closeQuietly(client);
+            }
+        }
+        finally
+        {
+            System.clearProperty("znode.container.checkIntervalMs");
         }
     }
 
@@ -199,7 +306,7 @@ public abstract class TestInterProcessMutexBase extends BaseClassForTests
         }
         finally
         {
-            client.close();
+            CloseableUtils.closeQuietly(client);
         }
     }
 
@@ -230,9 +337,9 @@ public abstract class TestInterProcessMutexBase extends BaseClassForTests
                             {
                                 semaphore.acquire();
                                 mutex.acquire();
-                                Assert.assertTrue(hasLock.compareAndSet(false, true));
                                 try
                                 {
+                                    assertTrue(hasLock.compareAndSet(false, true));
                                     if ( isFirst.compareAndSet(true, false) )
                                     {
                                         semaphore.release(THREAD_QTY - 1);
@@ -265,7 +372,7 @@ public abstract class TestInterProcessMutexBase extends BaseClassForTests
         }
         finally
         {
-            client.close();
+            CloseableUtils.closeQuietly(client);
         }
     }
 
@@ -287,11 +394,11 @@ public abstract class TestInterProcessMutexBase extends BaseClassForTests
                         @Override
                         public Object call() throws Exception
                         {
-                            Assert.assertTrue(countLatchForBar.await(10, TimeUnit.SECONDS));
+                            assertTrue(countLatchForBar.await(10, TimeUnit.SECONDS));
                             try
                             {
                                 mutex.acquire(10, TimeUnit.SECONDS);
-                                Assert.fail();
+                                fail();
                             }
                             catch ( Exception e )
                             {
@@ -307,11 +414,11 @@ public abstract class TestInterProcessMutexBase extends BaseClassForTests
                 );
 
             foo(mutex);
-            Assert.assertFalse(mutex.isAcquiredInThisProcess());
+            assertFalse(mutex.isAcquiredInThisProcess());
         }
         finally
         {
-            client.close();
+            CloseableUtils.closeQuietly(client);
         }
     }
 
@@ -324,43 +431,43 @@ public abstract class TestInterProcessMutexBase extends BaseClassForTests
         {
             InterProcessLock mutex = makeLock(client);
             foo(mutex);
-            Assert.assertFalse(mutex.isAcquiredInThisProcess());
+            assertFalse(mutex.isAcquiredInThisProcess());
         }
         finally
         {
-            client.close();
+            CloseableUtils.closeQuietly(client);
         }
     }
 
     private void foo(InterProcessLock mutex) throws Exception
     {
         mutex.acquire(10, TimeUnit.SECONDS);
-        Assert.assertTrue(mutex.isAcquiredInThisProcess());
+        assertTrue(mutex.isAcquiredInThisProcess());
         bar(mutex);
-        Assert.assertTrue(mutex.isAcquiredInThisProcess());
+        assertTrue(mutex.isAcquiredInThisProcess());
         mutex.release();
     }
 
     private void bar(InterProcessLock mutex) throws Exception
     {
         mutex.acquire(10, TimeUnit.SECONDS);
-        Assert.assertTrue(mutex.isAcquiredInThisProcess());
+        assertTrue(mutex.isAcquiredInThisProcess());
         if ( countLatchForBar != null )
         {
             countLatchForBar.countDown();
             waitLatchForBar.await(10, TimeUnit.SECONDS);
         }
         snafu(mutex);
-        Assert.assertTrue(mutex.isAcquiredInThisProcess());
+        assertTrue(mutex.isAcquiredInThisProcess());
         mutex.release();
     }
 
     private void snafu(InterProcessLock mutex) throws Exception
     {
         mutex.acquire(10, TimeUnit.SECONDS);
-        Assert.assertTrue(mutex.isAcquiredInThisProcess());
+        assertTrue(mutex.isAcquiredInThisProcess());
         mutex.release();
-        Assert.assertTrue(mutex.isAcquiredInThisProcess());
+        assertTrue(mutex.isAcquiredInThisProcess());
     }
 
     @Test
@@ -434,13 +541,13 @@ public abstract class TestInterProcessMutexBase extends BaseClassForTests
             while ( !mutexForClient1.isAcquiredInThisProcess() && !mutexForClient2.isAcquiredInThisProcess() )
             {
                 Thread.sleep(1000);
-                Assert.assertFalse(future1.isDone() && future2.isDone());
+                assertFalse(future1.isDone() && future2.isDone());
             }
 
-            Assert.assertTrue(mutexForClient1.isAcquiredInThisProcess() != mutexForClient2.isAcquiredInThisProcess());
+            assertTrue(mutexForClient1.isAcquiredInThisProcess() != mutexForClient2.isAcquiredInThisProcess());
             Thread.sleep(1000);
-            Assert.assertTrue(mutexForClient1.isAcquiredInThisProcess() || mutexForClient2.isAcquiredInThisProcess());
-            Assert.assertTrue(mutexForClient1.isAcquiredInThisProcess() != mutexForClient2.isAcquiredInThisProcess());
+            assertTrue(mutexForClient1.isAcquiredInThisProcess() || mutexForClient2.isAcquiredInThisProcess());
+            assertTrue(mutexForClient1.isAcquiredInThisProcess() != mutexForClient2.isAcquiredInThisProcess());
 
             Exception exception = exceptionRef.get();
             if ( exception != null )
@@ -451,15 +558,18 @@ public abstract class TestInterProcessMutexBase extends BaseClassForTests
             if ( mutexForClient1.isAcquiredInThisProcess() )
             {
                 latchForClient1.countDown();
-                Assert.assertTrue(acquiredLatchForClient2.await(10, TimeUnit.SECONDS));
-                Assert.assertTrue(mutexForClient2.isAcquiredInThisProcess());
+                assertTrue(acquiredLatchForClient2.await(10, TimeUnit.SECONDS));
+                assertTrue(mutexForClient2.isAcquiredInThisProcess());
             }
             else
             {
                 latchForClient2.countDown();
-                Assert.assertTrue(acquiredLatchForClient1.await(10, TimeUnit.SECONDS));
-                Assert.assertTrue(mutexForClient1.isAcquiredInThisProcess());
+                assertTrue(acquiredLatchForClient1.await(10, TimeUnit.SECONDS));
+                assertTrue(mutexForClient1.isAcquiredInThisProcess());
             }
+
+            future1.get();
+            future2.get();
         }
         finally
         {
