@@ -19,19 +19,33 @@
 
 package org.apache.curator.framework.recipes.locks;
 
+import static org.junit.jupiter.api.Assertions.assertEquals;
+import static org.junit.jupiter.api.Assertions.assertNotNull;
+import static org.junit.jupiter.api.Assertions.assertNull;
+import static org.junit.jupiter.api.Assertions.assertTrue;
+import static org.junit.jupiter.api.Assertions.fail;
 import com.google.common.collect.Lists;
-import org.apache.curator.test.BaseClassForTests;
-import org.apache.curator.utils.CloseableUtils;
+import com.google.common.collect.Queues;
 import org.apache.curator.framework.CuratorFramework;
 import org.apache.curator.framework.CuratorFrameworkFactory;
+import org.apache.curator.framework.api.CuratorWatcher;
+import org.apache.curator.framework.imps.TestCleanState;
 import org.apache.curator.framework.recipes.shared.SharedCount;
+import org.apache.curator.framework.state.ConnectionState;
+import org.apache.curator.framework.state.ConnectionStateListener;
 import org.apache.curator.retry.RetryOneTime;
+import org.apache.curator.test.BaseClassForTests;
 import org.apache.curator.test.Timing;
-import org.testng.Assert;
-import org.testng.annotations.Test;
+import org.apache.curator.utils.CloseableUtils;
+import org.apache.zookeeper.KeeperException;
+import org.apache.zookeeper.WatchedEvent;
+import org.apache.zookeeper.Watcher;
+import org.junit.jupiter.api.Test;
+
 import java.util.Collection;
 import java.util.List;
 import java.util.Random;
+import java.util.concurrent.BlockingQueue;
 import java.util.concurrent.Callable;
 import java.util.concurrent.CountDownLatch;
 import java.util.concurrent.ExecutorCompletionService;
@@ -45,6 +59,125 @@ import java.util.concurrent.atomic.AtomicInteger;
 @SuppressWarnings({"SynchronizationOnLocalVariableOrMethodParameter"})
 public class TestInterProcessSemaphore extends BaseClassForTests
 {
+    @Test
+    public void testAcquireAfterLostServer() throws Exception
+    {
+        // CURATOR-335
+
+        final String SEMAPHORE_PATH = "/test";
+        final int MAX_SEMAPHORES = 1;
+        final int NUM_CLIENTS = 10;
+
+        ExecutorService executor = Executors.newFixedThreadPool(NUM_CLIENTS);
+
+        final Timing timing = new Timing();
+
+        final CuratorFramework client = CuratorFrameworkFactory.newClient(server.getConnectString(), timing.forWaiting().milliseconds(), timing.connection(), new RetryOneTime(1));  // long session time on purpose
+        try
+        {
+            client.start();
+
+            InterProcessSemaphoreV2.debugAcquireLatch = new CountDownLatch(1);  // cause one of the semaphores to create its node and then wait
+            InterProcessSemaphoreV2.debugFailedGetChildrenLatch = new CountDownLatch(1);    // semaphore will notify when getChildren() fails
+            final CountDownLatch isReadyLatch = new CountDownLatch(NUM_CLIENTS);
+            final BlockingQueue<Boolean> acquiredQueue = Queues.newLinkedBlockingQueue();
+            Runnable runner = new Runnable()
+            {
+                @Override
+                public void run()
+                {
+                    while ( !Thread.currentThread().isInterrupted() )
+                    {
+                        InterProcessSemaphoreV2 semaphore = new InterProcessSemaphoreV2(client, SEMAPHORE_PATH, MAX_SEMAPHORES);
+                        Lease lease = null;
+                        try
+                        {
+                            isReadyLatch.countDown();
+                            lease = semaphore.acquire();
+                            acquiredQueue.add(true);
+                            timing.sleepABit();
+                        }
+                        catch ( InterruptedException e )
+                        {
+                            Thread.currentThread().interrupt();
+                            break;
+                        }
+                        catch ( KeeperException e )
+                        {
+                            try
+                            {
+                                timing.sleepABit();
+                            }
+                            catch ( InterruptedException e2 )
+                            {
+                                Thread.currentThread().interrupt();
+                                break;
+                            }
+                        }
+                        catch ( Exception ignore )
+                        {
+                            // ignore
+                        }
+                        finally
+                        {
+                            if ( lease != null )
+                            {
+                                semaphore.returnLease(lease);
+                            }
+                        }
+                    }
+                }
+            };
+            for ( int i = 0; i < NUM_CLIENTS; ++i )
+            {
+                executor.execute(runner);
+            }
+            assertTrue(timing.awaitLatch(isReadyLatch));
+            timing.sleepABit();
+
+            final CountDownLatch lostLatch = new CountDownLatch(1);
+            final CountDownLatch restartedLatch = new CountDownLatch(1);
+            client.getConnectionStateListenable().addListener(new ConnectionStateListener()
+            {
+                @Override
+                public void stateChanged(CuratorFramework client, ConnectionState newState)
+                {
+                    if ( newState == ConnectionState.LOST )
+                    {
+                        lostLatch.countDown();
+                    }
+                    else if ( newState == ConnectionState.RECONNECTED  )
+                    {
+                        restartedLatch.countDown();
+                    }
+                }
+            });
+      
+            server.stop();
+            assertTrue(timing.multiple(1.25).awaitLatch(lostLatch));
+            InterProcessSemaphoreV2.debugAcquireLatch.countDown();  // the waiting semaphore proceeds to getChildren - which should fail
+            assertTrue(timing.awaitLatch(InterProcessSemaphoreV2.debugFailedGetChildrenLatch));  // wait until getChildren fails
+
+            server.restart();
+
+            assertTrue(timing.awaitLatch(restartedLatch));
+            for ( int i = 0; i < NUM_CLIENTS; ++i )
+            {
+                // acquires should continue as normal after server restart
+                Boolean polled = acquiredQueue.poll(timing.forWaiting().milliseconds(), TimeUnit.MILLISECONDS);
+                if ( (polled == null) || !polled )
+                {
+                    fail("Semaphores not reacquired after restart");
+                }
+            }
+        }
+        finally
+        {
+            executor.shutdownNow();
+            CloseableUtils.closeQuietly(client);
+        }
+    }
+
     @Test
     public void testThreadedLeaseIncrease() throws Exception
     {
@@ -71,10 +204,10 @@ public class TestInterProcessSemaphore extends BaseClassForTests
                         public Object call() throws Exception
                         {
                             Lease lease = semaphore.acquire(timing.seconds(), TimeUnit.SECONDS);
-                            Assert.assertNotNull(lease);
+                            assertNotNull(lease);
                             latch1.countDown();
                             lease = semaphore.acquire(timing.forWaiting().seconds(), TimeUnit.SECONDS);
-                            Assert.assertNotNull(lease);
+                            assertNotNull(lease);
                             latch2.countDown();
                             return null;
                         }
@@ -87,12 +220,12 @@ public class TestInterProcessSemaphore extends BaseClassForTests
                         @Override
                         public Object call() throws Exception
                         {
-                            Assert.assertTrue(latch1.await(timing.forWaiting().seconds(), TimeUnit.SECONDS));
+                            assertTrue(latch1.await(timing.forWaiting().seconds(), TimeUnit.SECONDS));
                             timing.sleepABit(); // make sure second acquire is waiting
-                            Assert.assertTrue(count.trySetCount(2));
+                            assertTrue(count.trySetCount(2));
                             //Make sure second acquire takes less than full waiting time:
                             timing.sleepABit();
-                            Assert.assertTrue(latch2.await(0, TimeUnit.SECONDS));
+                            assertTrue(latch2.await(0, TimeUnit.SECONDS));
                             return null;
                         }
                     }
@@ -100,10 +233,12 @@ public class TestInterProcessSemaphore extends BaseClassForTests
 
             future1.get();
             future2.get();
+
+            count.close();
         }
         finally
         {
-            CloseableUtils.closeQuietly(client);
+            TestCleanState.closeAndTestClean(client);
         }
     }
 
@@ -127,21 +262,21 @@ public class TestInterProcessSemaphore extends BaseClassForTests
             semaphore2 = new InterProcessSemaphoreV2(client2, "/test", 1);
 
             Lease lease = semaphore2.acquire(timing.forWaiting().seconds(), TimeUnit.SECONDS);
-            Assert.assertNotNull(lease);
+            assertNotNull(lease);
             lease.close();
 
             lease = semaphore1.acquire(10, TimeUnit.SECONDS);
-            Assert.assertNotNull(lease);
+            assertNotNull(lease);
 
             client1.close();    // should release any held leases
             client1 = null;
 
-            Assert.assertNotNull(semaphore2.acquire(timing.forWaiting().seconds(), TimeUnit.SECONDS));
+            assertNotNull(semaphore2.acquire(timing.forWaiting().seconds(), TimeUnit.SECONDS));
         }
         finally
         {
-            CloseableUtils.closeQuietly(client1);
-            CloseableUtils.closeQuietly(client2);
+            TestCleanState.closeAndTestClean(client1);
+            TestCleanState.closeAndTestClean(client2);
         }
     }
 
@@ -193,10 +328,10 @@ public class TestInterProcessSemaphore extends BaseClassForTests
                                                 thisQty = (available.get() > 1) ? (random.nextInt(available.get()) + 1) : 1;
 
                                                 available.addAndGet(-1 * thisQty);
-                                                Assert.assertTrue(available.get() >= 0);
+                                                assertTrue(available.get() >= 0);
                                             }
                                             Collection<Lease> leases = semaphore.acquire(thisQty, timing.forWaiting().seconds(), TimeUnit.SECONDS);
-                                            Assert.assertNotNull(leases);
+                                            assertNotNull(leases);
                                             try
                                             {
                                                 synchronized(counter)
@@ -226,7 +361,7 @@ public class TestInterProcessSemaphore extends BaseClassForTests
                                     }
                                     finally
                                     {
-                                        client.close();
+                                        TestCleanState.closeAndTestClean(client);
                                     }
                                     return null;
                                 }
@@ -242,9 +377,9 @@ public class TestInterProcessSemaphore extends BaseClassForTests
 
         synchronized(counter)
         {
-            Assert.assertTrue(counter.currentCount == 0);
-            Assert.assertTrue(counter.maxCount > 0);
-            Assert.assertTrue(counter.maxCount <= SESSION_MAX);
+            assertTrue(counter.currentCount == 0);
+            assertTrue(counter.maxCount > 0);
+            assertTrue(counter.maxCount <= SESSION_MAX);
             System.out.println(counter.maxCount);
         }
     }
@@ -276,7 +411,7 @@ public class TestInterProcessSemaphore extends BaseClassForTests
                             {
                                 InterProcessSemaphoreV2 semaphore = new InterProcessSemaphoreV2(client, "/test", MAX);
                                 Lease lease = semaphore.acquire(timing.forWaiting().seconds(), TimeUnit.SECONDS);
-                                Assert.assertNotNull(lease);
+                                assertNotNull(lease);
                                 uses.incrementAndGet();
                                 try
                                 {
@@ -299,7 +434,7 @@ public class TestInterProcessSemaphore extends BaseClassForTests
                             }
                             finally
                             {
-                                client.close();
+                                TestCleanState.closeAndTestClean(client);
                             }
                             return null;
                         }
@@ -313,8 +448,8 @@ public class TestInterProcessSemaphore extends BaseClassForTests
             f.get();
         }
 
-        Assert.assertEquals(uses.get(), CLIENT_QTY);
-        Assert.assertEquals(maxLeases.get(), MAX);
+        assertEquals(uses.get(), CLIENT_QTY);
+        assertEquals(maxLeases.get(), MAX);
     }
 
     @Test
@@ -393,15 +528,15 @@ public class TestInterProcessSemaphore extends BaseClassForTests
             timing.sleepABit();
             synchronized(counter)
             {
-                Assert.assertTrue(counter.currentCount == 0);
-                Assert.assertTrue(counter.maxCount > 0);
-                Assert.assertTrue(counter.maxCount <= MAX_LEASES);
+                assertTrue(counter.currentCount == 0);
+                assertTrue(counter.maxCount > 0);
+                assertTrue(counter.maxCount <= MAX_LEASES);
                 System.out.println(counter.maxCount);
             }
         }
         finally
         {
-            client.close();
+            TestCleanState.closeAndTestClean(client);
         }
     }
 
@@ -441,11 +576,11 @@ public class TestInterProcessSemaphore extends BaseClassForTests
                     );
             }
             service.shutdown();
-            Assert.assertTrue(service.awaitTermination(10, TimeUnit.SECONDS));
+            assertTrue(service.awaitTermination(10, TimeUnit.SECONDS));
         }
         finally
         {
-            client.close();
+            TestCleanState.closeAndTestClean(client);
         }
     }
 
@@ -458,12 +593,12 @@ public class TestInterProcessSemaphore extends BaseClassForTests
         try
         {
             InterProcessSemaphoreV2 semaphore = new InterProcessSemaphoreV2(client, "/test", 1);
-            Assert.assertNotNull(semaphore.acquire(timing.forWaiting().seconds(), TimeUnit.SECONDS));
-            Assert.assertNull(semaphore.acquire(timing.forWaiting().seconds(), TimeUnit.SECONDS));
+            assertNotNull(semaphore.acquire(timing.forWaiting().seconds(), TimeUnit.SECONDS));
+            assertNull(semaphore.acquire(timing.forWaiting().seconds(), TimeUnit.SECONDS));
         }
         finally
         {
-            client.close();
+            TestCleanState.closeAndTestClean(client);
         }
     }
 
@@ -482,16 +617,16 @@ public class TestInterProcessSemaphore extends BaseClassForTests
             {
                 InterProcessSemaphoreV2 semaphore = new InterProcessSemaphoreV2(client, "/test", MAX_LEASES);
                 Lease lease = semaphore.acquire(timing.forWaiting().seconds(), TimeUnit.SECONDS);
-                Assert.assertNotNull(lease);
+                assertNotNull(lease);
                 leases.add(lease);
             }
 
             InterProcessSemaphoreV2 semaphore = new InterProcessSemaphoreV2(client, "/test", MAX_LEASES);
             Lease lease = semaphore.acquire(timing.forWaiting().seconds(), TimeUnit.SECONDS);
-            Assert.assertNull(lease);
+            assertNull(lease);
 
             leases.remove(0).close();
-            Assert.assertNotNull(semaphore.acquire(timing.forWaiting().seconds(), TimeUnit.SECONDS));
+            assertNotNull(semaphore.acquire(timing.forWaiting().seconds(), TimeUnit.SECONDS));
         }
         finally
         {
@@ -499,7 +634,7 @@ public class TestInterProcessSemaphore extends BaseClassForTests
             {
                 CloseableUtils.closeQuietly(l);
             }
-            CloseableUtils.closeQuietly(client);
+            TestCleanState.closeAndTestClean(client);
         }
     }
 
@@ -520,7 +655,7 @@ public class TestInterProcessSemaphore extends BaseClassForTests
                 leases.add(semaphore.acquire());
             }
 
-            Assert.assertEquals(semaphore.getParticipantNodes().size(), LEASES);
+            assertEquals(semaphore.getParticipantNodes().size(), LEASES);
         }
         finally
         {
@@ -528,7 +663,125 @@ public class TestInterProcessSemaphore extends BaseClassForTests
             {
                 CloseableUtils.closeQuietly(l);
             }
-            CloseableUtils.closeQuietly(client);
+            TestCleanState.closeAndTestClean(client);
+        }
+    }
+
+    @Test
+    public void testNoOrphanedNodes() throws Exception
+    {
+        final Timing timing = new Timing();
+        final ExecutorService executor = Executors.newFixedThreadPool(1);
+        CuratorFramework client = CuratorFrameworkFactory.newClient(server.getConnectString(), timing.session(), timing.connection(), new RetryOneTime(1));
+        client.start();
+        try
+        {
+            final InterProcessSemaphoreV2 semaphore = new InterProcessSemaphoreV2(client, "/test", 1);
+            Lease lease = semaphore.acquire(timing.forWaiting().seconds(), TimeUnit.SECONDS);
+            assertNotNull(lease);
+            final List<String> childNodes = client.getChildren().forPath("/test/leases");
+            assertEquals(childNodes.size(), 1);
+
+            final CountDownLatch nodeCreatedLatch = new CountDownLatch(1);
+            client.getChildren().usingWatcher(new CuratorWatcher()
+            {
+                @Override
+                public void process(WatchedEvent event) throws Exception
+                {
+                    if ( event.getType() == Watcher.Event.EventType.NodeCreated )
+                    {
+                        nodeCreatedLatch.countDown();
+                    }
+                }
+            }).forPath("/test/leases");
+
+            final Future<Lease> leaseFuture = executor.submit(new Callable<Lease>()
+            {
+                @Override
+                public Lease call() throws Exception
+                {
+                    return semaphore.acquire(timing.forWaiting().multiple(2).seconds(), TimeUnit.SECONDS);
+                }
+            });
+
+            // wait for second lease to create its node
+            timing.awaitLatch(nodeCreatedLatch);
+            String newNode = null;
+            for ( String c : client.getChildren().forPath("/test/leases") )
+            {
+                if ( !childNodes.contains(c) )
+                {
+                    newNode = c;
+                }
+            }
+            assertNotNull(newNode);
+
+            // delete the ephemeral node to trigger a retry
+            client.delete().forPath("/test/leases/" + newNode);
+
+            // release first lease so second one can be acquired
+            lease.close();
+            lease = leaseFuture.get();
+            assertNotNull(lease);
+            lease.close();
+            assertEquals(client.getChildren().forPath("/test/leases").size(), 0);
+
+            // no more lease exist. must be possible to acquire a new one
+            assertNotNull(semaphore.acquire(timing.forWaiting().seconds(), TimeUnit.SECONDS));
+        }
+        finally
+        {
+            executor.shutdownNow();
+            TestCleanState.closeAndTestClean(client);
+        }
+    }
+    
+    @Test
+    public void testInterruptAcquire() throws Exception
+    {
+        // CURATOR-462
+        final Timing timing = new Timing();
+        CuratorFramework client = CuratorFrameworkFactory.newClient(server.getConnectString(), timing.session(), timing.connection(), new RetryOneTime(1));
+        client.start();
+        try
+        {
+            final InterProcessSemaphoreV2 s1 = new InterProcessSemaphoreV2(client, "/test", 1);
+            final InterProcessSemaphoreV2 s2 = new InterProcessSemaphoreV2(client, "/test", 1);
+            final InterProcessSemaphoreV2 s3 = new InterProcessSemaphoreV2(client, "/test", 1);
+            
+            final CountDownLatch debugWaitLatch = s2.debugWaitLatch = new CountDownLatch(1);
+            
+            // Acquire exclusive semaphore
+            Lease lease = s1.acquire(timing.forWaiting().seconds(), TimeUnit.SECONDS);
+            assertNotNull(lease);
+            
+            // Queue up another semaphore on the same path
+            Future<Object> handle = Executors.newSingleThreadExecutor().submit(new Callable<Object>() {
+                
+                @Override
+                public Object call() throws Exception {
+                    s2.acquire();
+                    return null;
+                }
+            });
+
+            // Wait until second lease is created and the wait is started for it to become active
+            assertTrue(timing.awaitLatch(debugWaitLatch));
+            
+            // Interrupt the wait
+            handle.cancel(true);
+            
+            // Assert that the second lease is gone
+            timing.sleepABit();
+            assertEquals(client.getChildren().forPath("/test/leases").size(), 1);
+            
+            // Assert that after closing the first (current) semaphore, we can acquire a new one
+            s1.returnLease(lease);
+            assertNotNull(s3.acquire(timing.forWaiting().seconds(), TimeUnit.SECONDS));
+        }
+        finally
+        {
+            TestCleanState.closeAndTestClean(client);
         }
     }
 }
