@@ -18,17 +18,16 @@
  */
 package org.apache.curator.test;
 
-import org.apache.zookeeper.server.quorum.QuorumPeer;
-import org.apache.zookeeper.server.quorum.QuorumPeerConfig;
 import java.io.Closeable;
+import org.apache.zookeeper.server.quorum.QuorumPeerConfig;
 
 public interface ZooKeeperMainFace extends Closeable
 {
-    public void runFromConfig(QuorumPeerConfig config) throws Exception;
+    void configure(QuorumConfigBuilder config, int instanceIndex) throws Exception;
 
-    public void blockUntilStarted() throws Exception;
+    void start();
 
-    public void kill();
+    void kill();
 
-    public QuorumPeer getQuorumPeer();
+    QuorumPeerConfig getConfig() throws Exception;
 }
